# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Preconfigured checks for Python code.

These checks assume that they are running in a preconfigured Python environment.
"""

import json
import logging
import os
from pathlib import Path
import subprocess
import sys
from typing import Optional

try:
    import pw_presubmit
except ImportError:
    # Append the pw_presubmit package path to the module search path to allow
    # running this module without installing the pw_presubmit package.
    sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
    import pw_presubmit

from pw_env_setup import python_packages
from pw_presubmit import (
    build,
    call,
    Check,
    filter_paths,
    PresubmitContext,
    PresubmitFailure,
)

_LOG = logging.getLogger(__name__)

_PYTHON_EXTENSIONS = ('.py', '.gn', '.gni')

_PYTHON_IS_3_9_OR_HIGHER = sys.version_info >= (
    3,
    9,
)


@filter_paths(endswith=_PYTHON_EXTENSIONS)
def gn_python_check(ctx: PresubmitContext):
    build.gn_gen(ctx)
    build.ninja(ctx, 'python.tests', 'python.lint')


def _transform_lcov_file_paths(lcov_file: Path, repo_root: Path) -> str:
    """Modify file paths in an lcov file to be relative to the repo root.

    See `man geninfo` for info on the lcov format."""

    lcov_input = lcov_file.read_text()
    lcov_output = ''

    if not _PYTHON_IS_3_9_OR_HIGHER:
        return lcov_input

    for line in lcov_input.splitlines():
        if not line.startswith('SF:'):
            lcov_output += line + '\n'
            continue

        # Get the file path after SF:
        file_string = line[3:].rstrip()
        source_file_path = Path(file_string)

        # TODO(b/248257406) Remove once we drop support for Python 3.8.
        def is_relative_to(path: Path, other: Path) -> bool:
            try:
                path.relative_to(other)
                return True
            except ValueError:
                return False

        # Attempt to map a generated Python package source file to the root
        # source tree.
        # pylint: disable=no-member
        if not is_relative_to(
            source_file_path, repo_root  # type: ignore[attr-defined]
        ):
            # pylint: enable=no-member
            source_file_path = repo_root / str(source_file_path).replace(
                'python/gen/', ''
            ).replace('py.generated_python_package/', '')

        # If mapping fails don't modify this line.
        # pylint: disable=no-member
        if not is_relative_to(
            source_file_path, repo_root  # type: ignore[attr-defined]
        ):
            # pylint: enable=no-member
            lcov_output += line + '\n'
            continue

        source_file_path = source_file_path.relative_to(repo_root)
        lcov_output += f'SF:{source_file_path}\n'

    return lcov_output


@filter_paths(endswith=_PYTHON_EXTENSIONS)
def gn_python_test_coverage(ctx: PresubmitContext):
    """Run Python tests with coverage and create reports."""
    build.gn_gen(ctx, pw_build_PYTHON_TEST_COVERAGE=True)
    build.ninja(ctx, 'python.tests')

    # Find coverage data files
    coverage_data_files = list(ctx.output_dir.glob('**/*.coverage'))
    if not coverage_data_files:
        return

    # Merge coverage data files to out/.coverage
    call(
        'coverage',
        'combine',
        # Leave existing coverage files in place; by default they are deleted.
        '--keep',
        *coverage_data_files,
        cwd=ctx.output_dir,
    )
    combined_data_file = ctx.output_dir / '.coverage'
    _LOG.info('Coverage data saved to: %s', combined_data_file.resolve())

    # Always ignore generated proto python and setup.py files.
    coverage_omit_patterns = '--omit=*_pb2.py,*/setup.py'

    # Output coverage percentage summary to the terminal of changed files.
    changed_python_files = list(
        str(p) for p in ctx.paths if str(p).endswith('.py')
    )
    report_args = [
        'coverage',
        'report',
        '--ignore-errors',
        coverage_omit_patterns,
    ]
    report_args += changed_python_files
    subprocess.run(report_args, check=False, cwd=ctx.output_dir)

    # Generate a json report
    call('coverage', 'lcov', coverage_omit_patterns, cwd=ctx.output_dir)
    lcov_data_file = ctx.output_dir / 'coverage.lcov'
    lcov_data_file.write_text(
        _transform_lcov_file_paths(lcov_data_file, repo_root=ctx.root)
    )
    _LOG.info('Coverage lcov saved to: %s', lcov_data_file.resolve())

    # Generate an html report
    call('coverage', 'html', coverage_omit_patterns, cwd=ctx.output_dir)
    html_report = ctx.output_dir / 'htmlcov' / 'index.html'
    _LOG.info('Coverage html report saved to: %s', html_report.resolve())


@filter_paths(endswith=_PYTHON_EXTENSIONS + ('.pylintrc',))
def gn_python_lint(ctx: pw_presubmit.PresubmitContext) -> None:
    build.gn_gen(ctx)
    build.ninja(ctx, 'python.lint')


@Check
def check_python_versions(ctx: PresubmitContext):
    """Checks that the list of installed packages is as expected."""

    build.gn_gen(ctx)
    constraint_file: Optional[str] = None
    requirement_file: Optional[str] = None
    try:
        for arg in build.get_gn_args(ctx.output_dir):
            if arg['name'] == 'pw_build_PIP_CONSTRAINTS':
                constraint_file = json.loads(arg['current']['value'])[0].strip(
                    '/'
                )
            if arg['name'] == 'pw_build_PIP_REQUIREMENTS':
                requirement_file = json.loads(arg['current']['value'])[0].strip(
                    '/'
                )
    except json.JSONDecodeError:
        _LOG.warning('failed to parse GN args json')
        return

    if not constraint_file:
        _LOG.warning('could not find pw_build_PIP_CONSTRAINTS GN arg')
        return
    ignored_requirements_arg = None
    if requirement_file:
        ignored_requirements_arg = [(ctx.root / requirement_file)]

    if (
        python_packages.diff(
            expected=(ctx.root / constraint_file),
            ignore_requirements_file=ignored_requirements_arg,
        )
        != 0
    ):
        raise PresubmitFailure



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2021 The Pigweed Authors

Licensed under the Apache License, Version 2.0 (the "License"); you may not
use this file except in compliance with the License. You may obtain a copy of
the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""
import builtins
import google.protobuf.any_pb2
import google.protobuf.descriptor
import google.protobuf.internal.enum_type_wrapper
import google.protobuf.message
import sys
import typing

if sys.version_info >= (3, 10):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class BundledUpdateState(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Enum:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _EnumEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[BundledUpdateState._Enum.ValueType], builtins.type):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNKNOWN: BundledUpdateState._Enum.ValueType  # 0
        """Not an expected state in the OTA system; only for proto."""
        INACTIVE: BundledUpdateState._Enum.ValueType  # 1
        """Valid methods in this state: Start()

        Transition:
          Start() succeeds --> TRANSFERRING
          Start() fails    --> FINISHED
        """
        TRANSFERRING: BundledUpdateState._Enum.ValueType  # 2
        """Valid methods in this state: GetStatus(), Abort().

        Transitions:
          Transfer completes --> TRANSFERRED
          Transfer fails     --> FINISHED
          Abort() called     --> ABORTING
        """
        TRANSFERRED: BundledUpdateState._Enum.ValueType  # 3
        """Valid methods in this state: GetStatus(), Abort(), Verify().

        Transitions:
          Verify() called    --> VERIFYING
          Apply() called     --> VERIFYING
          Abort() called     --> ABORTING
        """
        VERIFYING: BundledUpdateState._Enum.ValueType  # 4
        """Valid methods in this state: GetStatus(), Abort().

        Transitions:
          Verifying finished --> VERIFIED
          Verifying failed   --> FINISHED
          Abort() called     --> ABORTING
        """
        VERIFIED: BundledUpdateState._Enum.ValueType  # 5
        """Valid methods in this state: GetStatus(), Abort(), Apply().

        Transitions:
          Apply() called --> APPLYING
          Abort() called --> ABORTING
        """
        APPLYING: BundledUpdateState._Enum.ValueType  # 6
        """Valid methods in this state: GetStatus().

        Transitions:
          Apply() finished --> FINISHED; may require persisting across a reboot.
          Apply() failed   --> FINISHED; with error set.
        """
        ABORTING: BundledUpdateState._Enum.ValueType  # 7
        """Valid methods in this state: GetStatus().

        Transitions:
          Abort finishes --> FINISHED
          Abort fails    --> FINISHED
        """
        FINISHED: BundledUpdateState._Enum.ValueType  # 8
        """Valid methods in this state: GetStatus(), Reset().

        Terminal state indicating a finished update; whether successful or
        not. Additional termination information available in completion_state
        and possibly note.

        Transitions:
          Reset() succeeds --> INACTIVE
          Reset() fails    --> FINISHED
        """

    class Enum(_Enum, metaclass=_EnumEnumTypeWrapper): ...
    UNKNOWN: BundledUpdateState.Enum.ValueType  # 0
    """Not an expected state in the OTA system; only for proto."""
    INACTIVE: BundledUpdateState.Enum.ValueType  # 1
    """Valid methods in this state: Start()

    Transition:
      Start() succeeds --> TRANSFERRING
      Start() fails    --> FINISHED
    """
    TRANSFERRING: BundledUpdateState.Enum.ValueType  # 2
    """Valid methods in this state: GetStatus(), Abort().

    Transitions:
      Transfer completes --> TRANSFERRED
      Transfer fails     --> FINISHED
      Abort() called     --> ABORTING
    """
    TRANSFERRED: BundledUpdateState.Enum.ValueType  # 3
    """Valid methods in this state: GetStatus(), Abort(), Verify().

    Transitions:
      Verify() called    --> VERIFYING
      Apply() called     --> VERIFYING
      Abort() called     --> ABORTING
    """
    VERIFYING: BundledUpdateState.Enum.ValueType  # 4
    """Valid methods in this state: GetStatus(), Abort().

    Transitions:
      Verifying finished --> VERIFIED
      Verifying failed   --> FINISHED
      Abort() called     --> ABORTING
    """
    VERIFIED: BundledUpdateState.Enum.ValueType  # 5
    """Valid methods in this state: GetStatus(), Abort(), Apply().

    Transitions:
      Apply() called --> APPLYING
      Abort() called --> ABORTING
    """
    APPLYING: BundledUpdateState.Enum.ValueType  # 6
    """Valid methods in this state: GetStatus().

    Transitions:
      Apply() finished --> FINISHED; may require persisting across a reboot.
      Apply() failed   --> FINISHED; with error set.
    """
    ABORTING: BundledUpdateState.Enum.ValueType  # 7
    """Valid methods in this state: GetStatus().

    Transitions:
      Abort finishes --> FINISHED
      Abort fails    --> FINISHED
    """
    FINISHED: BundledUpdateState.Enum.ValueType  # 8
    """Valid methods in this state: GetStatus(), Reset().

    Terminal state indicating a finished update; whether successful or
    not. Additional termination information available in completion_state
    and possibly note.

    Transitions:
      Reset() succeeds --> INACTIVE
      Reset() fails    --> FINISHED
    """

    def __init__(
        self,
    ) -> None: ...

global___BundledUpdateState = BundledUpdateState

class BundledUpdateResult(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    class _Enum:
        ValueType = typing.NewType("ValueType", builtins.int)
        V: typing_extensions.TypeAlias = ValueType

    class _EnumEnumTypeWrapper(google.protobuf.internal.enum_type_wrapper._EnumTypeWrapper[BundledUpdateResult._Enum.ValueType], builtins.type):  # noqa: F821
        DESCRIPTOR: google.protobuf.descriptor.EnumDescriptor
        UNKNOWN: BundledUpdateResult._Enum.ValueType  # 0
        SUCCESS: BundledUpdateResult._Enum.ValueType  # 1
        UNKNOWN_ERROR: BundledUpdateResult._Enum.ValueType  # 2
        ABORTED: BundledUpdateResult._Enum.ValueType  # 3
        TRANSFER_FAILED: BundledUpdateResult._Enum.ValueType  # 4
        VERIFY_FAILED: BundledUpdateResult._Enum.ValueType  # 5
        APPLY_FAILED: BundledUpdateResult._Enum.ValueType  # 6

    class Enum(_Enum, metaclass=_EnumEnumTypeWrapper): ...
    UNKNOWN: BundledUpdateResult.Enum.ValueType  # 0
    SUCCESS: BundledUpdateResult.Enum.ValueType  # 1
    UNKNOWN_ERROR: BundledUpdateResult.Enum.ValueType  # 2
    ABORTED: BundledUpdateResult.Enum.ValueType  # 3
    TRANSFER_FAILED: BundledUpdateResult.Enum.ValueType  # 4
    VERIFY_FAILED: BundledUpdateResult.Enum.ValueType  # 5
    APPLY_FAILED: BundledUpdateResult.Enum.ValueType  # 6

    def __init__(
        self,
    ) -> None: ...

global___BundledUpdateResult = BundledUpdateResult

class BundledUpdateStatus(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    STATE_FIELD_NUMBER: builtins.int
    RESULT_FIELD_NUMBER: builtins.int
    CURRENT_STATE_PROGRESS_HUNDRETH_PERCENT_FIELD_NUMBER: builtins.int
    TRANSFER_ID_FIELD_NUMBER: builtins.int
    BUNDLE_FILENAME_FIELD_NUMBER: builtins.int
    NOTE_FIELD_NUMBER: builtins.int
    EXTENDED_STATUS_FIELD_NUMBER: builtins.int
    state: global___BundledUpdateState.Enum.ValueType
    result: global___BundledUpdateResult.Enum.ValueType
    current_state_progress_hundreth_percent: builtins.int
    """This is the percentage of estimated progress for the current update
    state in hundreths of a percent. (e.g. 5.00% = 500u)
    """
    transfer_id: builtins.int
    """If present, the active transfer ID for the update."""
    bundle_filename: builtins.str
    """The name of the update bundle. Not present when in INACTIVE state. This is
    useful for enabling resuming of transfers across reboots or disconnects,
    without transferring an expensive manifest.
    """
    note: builtins.bytes
    """Additional information related to the state may be provided here.
    Examples: "Failed verifying: ml_model.bin", "Flash partition couldn't be
    acquired and was busy", etc. Can provide more granular information than
    just the completion result.
    """
    @property
    def extended_status(self) -> google.protobuf.any_pb2.Any:
        """Custom application data."""
    def __init__(
        self,
        *,
        state: global___BundledUpdateState.Enum.ValueType = ...,
        result: global___BundledUpdateResult.Enum.ValueType | None = ...,
        current_state_progress_hundreth_percent: builtins.int | None = ...,
        transfer_id: builtins.int | None = ...,
        bundle_filename: builtins.str | None = ...,
        note: builtins.bytes | None = ...,
        extended_status: google.protobuf.any_pb2.Any | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_bundle_filename", b"_bundle_filename", "_current_state_progress_hundreth_percent", b"_current_state_progress_hundreth_percent", "_extended_status", b"_extended_status", "_note", b"_note", "_result", b"_result", "_transfer_id", b"_transfer_id", "bundle_filename", b"bundle_filename", "current_state_progress_hundreth_percent", b"current_state_progress_hundreth_percent", "extended_status", b"extended_status", "note", b"note", "result", b"result", "transfer_id", b"transfer_id"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_bundle_filename", b"_bundle_filename", "_current_state_progress_hundreth_percent", b"_current_state_progress_hundreth_percent", "_extended_status", b"_extended_status", "_note", b"_note", "_result", b"_result", "_transfer_id", b"_transfer_id", "bundle_filename", b"bundle_filename", "current_state_progress_hundreth_percent", b"current_state_progress_hundreth_percent", "extended_status", b"extended_status", "note", b"note", "result", b"result", "state", b"state", "transfer_id", b"transfer_id"]) -> None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_bundle_filename", b"_bundle_filename"]) -> typing_extensions.Literal["bundle_filename"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_current_state_progress_hundreth_percent", b"_current_state_progress_hundreth_percent"]) -> typing_extensions.Literal["current_state_progress_hundreth_percent"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_extended_status", b"_extended_status"]) -> typing_extensions.Literal["extended_status"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_note", b"_note"]) -> typing_extensions.Literal["note"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_result", b"_result"]) -> typing_extensions.Literal["result"] | None: ...
    @typing.overload
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_transfer_id", b"_transfer_id"]) -> typing_extensions.Literal["transfer_id"] | None: ...

global___BundledUpdateStatus = BundledUpdateStatus

class StartRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BUNDLE_FILENAME_FIELD_NUMBER: builtins.int
    bundle_filename: builtins.str
    """If present, the filename for the staged file. This should persist across
    reboots, and will be returned from GetStatus() until either the update
    applies or is aborted.
    """
    def __init__(
        self,
        *,
        bundle_filename: builtins.str | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["_bundle_filename", b"_bundle_filename", "bundle_filename", b"bundle_filename"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["_bundle_filename", b"_bundle_filename", "bundle_filename", b"bundle_filename"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["_bundle_filename", b"_bundle_filename"]) -> typing_extensions.Literal["bundle_filename"] | None: ...

global___StartRequest = StartRequest



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Defines a class used to write code to an output buffer."""

from typing import List


class OutputFile:
    """A buffer to which data is written.

    Example:

    ```
    output = Output("hello.c")
    output.write_line('int main(void) {')
    with output.indent():
        output.write_line('printf("Hello, world");')
        output.write_line('return 0;')
    output.write_line('}')

    print(output.content())
    ```

    Produces:
    ```
    int main(void) {
      printf("Hello, world");
      return 0;
    }
    ```
    """

    INDENT_WIDTH = 2

    def __init__(self, filename: str):
        self._filename: str = filename
        self._content: List[str] = []
        self._indentation: int = 0

    def write_line(self, line: str = '') -> None:
        if line:
            self._content.append(' ' * self._indentation)
            self._content.append(line)
        self._content.append('\n')

    def indent(
        self,
        amount: int = INDENT_WIDTH,
    ) -> 'OutputFile._IndentationContext':
        """Increases the indentation level of the output."""
        return self._IndentationContext(self, amount)

    def name(self) -> str:
        return self._filename

    def content(self) -> str:
        return ''.join(self._content)

    class _IndentationContext:
        """Context that increases the output's indentation when it is active."""

        def __init__(self, output: 'OutputFile', amount: int):
            self._output = output
            self._amount: int = amount

        def __enter__(self):
            self._output._indentation += self._amount

        def __exit__(self, typ, value, traceback):
            self._output._indentation -= self._amount



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Size reporting utilities."""

import argparse
import logging
from pathlib import Path
import sys
from typing import Iterable

from pw_bloat import bloat
from pw_bloat.label import DataSourceMap
from pw_bloat.label_output import BloatTableOutput
import pw_cli.log

_LOG = logging.getLogger(__name__)


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument(
        'binary',
        help='Path to the ELF file to analyze',
        metavar='BINARY',
        type=Path,
    )
    parser.add_argument(
        '-d',
        '--data-sources',
        help='Comma-separated list of Bloaty data sources to report',
        type=lambda s: s.split(','),
        default=('memoryregions,sections'),
    )
    parser.add_argument(
        '--diff',
        metavar='BINARY',
        help='Run a size diff against a base binary file',
        type=Path,
    )
    parser.add_argument(
        '-v',
        '--verbose',
        help=('Print all log messages ' '(only errors are printed by default)'),
        action='store_true',
    )

    return parser.parse_args()


def _run_size_report(
    elf: Path, data_sources: Iterable[str] = ()
) -> DataSourceMap:
    """Runs a size analysis on an ELF file, returning a pw_bloat size map.

    Returns:
        A size map of the labels in the binary under the provided data sources.

    Raises:
        NoMemoryRegions: The binary does not define bloat memory region symbols.
    """

    bloaty_tsv = bloat.memory_regions_size_report(
        elf, data_sources=data_sources, extra_args=('--tsv',)
    )

    return DataSourceMap.from_bloaty_tsv(bloaty_tsv)


def _no_memory_regions_error(elf: Path) -> None:
    _LOG.error('Executable %s does not define any bloat memory regions', elf)
    _LOG.error(
        'Refer to https://pigweed.dev/pw_bloat/#memoryregions-data-source'
    )
    _LOG.error('for information on how to configure them.')


def _single_binary_report(elf: Path, data_sources: Iterable[str] = ()) -> int:
    try:
        data_source_map = _run_size_report(elf, data_sources)
    except bloat.NoMemoryRegions:
        _no_memory_regions_error(elf)
        return 1

    print(BloatTableOutput(data_source_map).create_table())
    return 0


def _diff_report(
    target: Path, base: Path, data_sources: Iterable[str] = ()
) -> int:
    try:
        base_map = _run_size_report(base, data_sources)
        target_map = _run_size_report(target, data_sources)
    except bloat.NoMemoryRegions as err:
        _no_memory_regions_error(err.elf)
        return 1

    diff = target_map.diff(base_map)

    print(BloatTableOutput(diff).create_table())
    return 0


def main() -> int:
    """Run binary size reports."""

    args = _parse_args()

    if not args.verbose:
        pw_cli.log.set_all_loggers_minimum_level(logging.ERROR)

    if args.diff is not None:
        return _diff_report(
            args.binary, args.diff, data_sources=args.data_sources
        )

    return _single_binary_report(args.binary, data_sources=args.data_sources)


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3
# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Invoke clang-tidy.

Implements additional features compared to directly calling
clang-tidy:
  - add option `--source-exclude` to exclude matching sources from the
    clang-tidy analysis.
  - inputs the full compile command, with the cc binary name
  - TODO(henrichataing): infer platform options from the full compile command
"""

import argparse
import logging
from pathlib import Path
import re
import shlex
import subprocess
import sys
from typing import Iterable, List, Optional, Union

_LOG = logging.getLogger(__name__)


def _parse_args() -> argparse.Namespace:
    """Parses arguments for this script, splitting out the command to run."""

    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-v',
        '--verbose',
        action='store_true',
        help='Run clang_tidy with extra debug output.',
    )

    parser.add_argument(
        '--clang-tidy',
        default='clang-tidy',
        help='Path to clang-tidy executable.',
    )

    parser.add_argument(
        '--source-file',
        required=True,
        type=Path,
        help='Path to the source file to analyze with clang-tidy.',
    )
    parser.add_argument(
        '--source-root',
        required=True,
        type=Path,
        help=(
            'Path to the root source directory.'
            ' The relative path from the root directory is matched'
            ' against source filter rather than the absolute path.'
        ),
    )
    parser.add_argument(
        '--export-fixes',
        required=False,
        type=Path,
        help=(
            'YAML file to store suggested fixes in. The '
            'stored fixes can be applied to the input source '
            'code with clang-apply-replacements.'
        ),
    )

    parser.add_argument(
        '--source-exclude',
        default=[],
        action='append',
        type=str,
        help=(
            'Regular expressions matching the paths of'
            ' source files to be excluded from the'
            ' analysis.'
        ),
    )

    parser.add_argument(
        '--skip-include-path',
        default=[],
        action='append',
        type=str,
        help=(
            'Exclude include paths ending in these paths from clang-tidy. '
            'These paths are switched from -I to -isystem so clang-tidy '
            'ignores them.'
        ),
    )

    # Add a silent placeholder arg for everything that was left over.
    parser.add_argument(
        'extra_args', nargs=argparse.REMAINDER, help=argparse.SUPPRESS
    )

    parsed_args = parser.parse_args()

    if parsed_args.extra_args[0] != '--':
        parser.error('arguments not correctly split')
    parsed_args.extra_args = parsed_args.extra_args[1:]
    return parsed_args


def _filter_include_paths(
    args: Iterable[str], skip_include_paths: Iterable[str]
) -> Iterable[str]:
    filters = [f.rstrip('/') for f in skip_include_paths]

    for arg in args:
        if arg.startswith('-I'):
            path = Path(arg[2:]).as_posix()
            if any(path.endswith(f) or re.match(f, str(path)) for f in filters):
                yield '-isystem' + arg[2:]
                continue
        if arg.startswith('--sysroot'):
            path = Path(arg[9:]).as_posix()
            if any(path.endswith(f) or re.match(f, str(path)) for f in filters):
                yield '-isysroot' + arg[9:]
                continue

        yield arg


def run_clang_tidy(
    clang_tidy: str,
    verbose: bool,
    source_file: Path,
    export_fixes: Optional[Path],
    skip_include_path: List[str],
    extra_args: List[str],
) -> int:
    """Executes clang_tidy via subprocess. Returns true if no failures."""
    command: List[Union[str, Path]] = [clang_tidy, source_file, '--use-color']

    if not verbose:
        command.append('--quiet')

    if export_fixes is not None:
        command.extend(['--export-fixes', export_fixes])

    # Append extra compilation flags.  Extra args up to
    # END_OF_INVOKER are skipped.
    command.append('--')
    end_of_invoker = extra_args.index('END_OF_INVOKER')
    command.extend(
        _filter_include_paths(
            extra_args[end_of_invoker + 1 :], skip_include_path
        )
    )

    process = subprocess.run(
        command,
        stdout=subprocess.PIPE,
        # clang-tidy prints regular information on
        # stderr, even with the option --quiet.
        stderr=subprocess.PIPE,
    )
    if process.returncode != 0:
        _LOG.warning('%s', ' '.join(shlex.quote(str(arg)) for arg in command))

    if process.stdout:
        _LOG.warning(process.stdout.decode().strip())

    if process.stderr and process.returncode != 0:
        _LOG.error(process.stderr.decode().strip())

    return process.returncode


def main(
    verbose: bool,
    clang_tidy: str,
    source_file: Path,
    source_root: Path,
    export_fixes: Optional[Path],
    source_exclude: List[str],
    skip_include_path: List[str],
    extra_args: List[str],
) -> int:
    # Rebase the source file path on source_root.
    # If source_file is not relative to source_root (which may be the case for
    # generated files) stick with the original source_file.
    try:
        relative_source_file = source_file.relative_to(source_root)
    except ValueError:
        relative_source_file = source_file

    for pattern in source_exclude:
        if re.match(pattern, str(relative_source_file)):
            return 0

    source_file_path = source_file.resolve()
    export_fixes_path = (
        export_fixes.resolve() if export_fixes is not None else None
    )
    return run_clang_tidy(
        clang_tidy,
        verbose,
        source_file_path,
        export_fixes_path,
        skip_include_path,
        extra_args,
    )


if __name__ == '__main__':
    sys.exit(main(**vars(_parse_args())))



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Copyright 2020 The Pigweed Authors

Licensed under the Apache License, Version 2.0 (the "License"); you may not
use this file except in compliance with the License. You may obtain a copy of
the License at

    https://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
License for the specific language governing permissions and limitations under
the License.
"""
import builtins
import collections.abc
import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import sys

if sys.version_info >= (3, 8):
    import typing as typing_extensions
else:
    import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class Metric(google.protobuf.message.Message):
    """A metric, described by the name (path + name), and the value.

    This flattened representation, while more complicated than the obvious tree
    structure alternative, enables streaming metrics from the device in low
    memory or low compute situations.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    TOKEN_PATH_FIELD_NUMBER: builtins.int
    STRING_PATH_FIELD_NUMBER: builtins.int
    AS_FLOAT_FIELD_NUMBER: builtins.int
    AS_INT_FIELD_NUMBER: builtins.int
    @property
    def token_path(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]:
        """The token path from the root. The last token is the metric name, and
        previous tokens are the parent group names. This could be converted from
        the tokens into a string; for example the token path {0xfaff, 0xabcd}:

         - The group is 0xfaff (root, parent)
         - The metric is 0xabcd

        Given the token database, this might be converted into:

          /i2c_bus_1/failed_transactions

        Note: This uses a repeated fixed32 instead of a "Oneof" with the string
        path to reduce the encoded size. Using a repeated Oneof name { str,
        fixed32 } would cost approximately 6N bytes for N path elements, vs 2 + 4N
        bytes in the packed case.
        """
    @property
    def string_path(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """The string path from the root. Similar to token path, but with strings.
        Note: This is currently unsupported.
        """
    as_float: builtins.float
    as_int: builtins.int
    def __init__(
        self,
        *,
        token_path: collections.abc.Iterable[builtins.int] | None = ...,
        string_path: collections.abc.Iterable[builtins.str] | None = ...,
        as_float: builtins.float = ...,
        as_int: builtins.int = ...,
    ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["as_float", b"as_float", "as_int", b"as_int", "value", b"value"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["as_float", b"as_float", "as_int", b"as_int", "string_path", b"string_path", "token_path", b"token_path", "value", b"value"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["value", b"value"]) -> typing_extensions.Literal["as_float", "as_int"] | None: ...

global___Metric = Metric

class MetricRequest(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    METRICS_FIELD_NUMBER: builtins.int
    @property
    def metrics(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Metric]:
        """Metrics or the groups matched to the given paths are returned.  The intent
        is to support matching semantics, with at least subsetting to e.g. collect
        all the metrics in a group and its children. We may also implement
        wildcard matchers.

        Value fields in the metrics will be ignored, since this is a query.

        Note: This is currently unsupported.
        """
    def __init__(
        self,
        *,
        metrics: collections.abc.Iterable[global___Metric] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["metrics", b"metrics"]) -> None: ...

global___MetricRequest = MetricRequest

class MetricResponse(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    METRICS_FIELD_NUMBER: builtins.int
    @property
    def metrics(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Metric]: ...
    def __init__(
        self,
        *,
        metrics: collections.abc.Iterable[global___Metric] | None = ...,
    ) -> None: ...
    def ClearField(self, field_name: typing_extensions.Literal["metrics", b"metrics"]) -> None: ...

global___MetricResponse = MetricResponse



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Crate a venv."""

import argparse
import venv
from pathlib import Path


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        '--destination-dir',
        type=Path,
        required=True,
        help='Path to venv directory.',
    )
    return parser.parse_args()


def main(destination_dir: Path) -> None:
    if not destination_dir.is_dir():
        venv.create(destination_dir, symlinks=True, with_pip=True)


if __name__ == '__main__':
    main(**vars(_parse_args()))



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Check the formatting of TODOs."""

import logging
from pathlib import Path
import re
from typing import Iterable, Pattern, Sequence, Union

from pw_presubmit import PresubmitContext, filter_paths

_LOG: logging.Logger = logging.getLogger(__name__)

EXCLUDE: Sequence[str] = (
    # Metadata
    r'^docker/tag$',
    r'\byarn.lock$',
    # Data files
    r'\.bin$',
    r'\.csv$',
    r'\.elf$',
    r'\.gif$',
    r'\.jpg$',
    r'\.json$',
    r'\.png$',
    r'\.svg$',
    r'\.xml$',
)

# todo-check: disable
BUGS_ONLY = re.compile(r'\bTODO\(b/\d+(?:, ?b/\d+)*\).*\w')
BUGS_OR_USERNAMES = re.compile(
    r'\bTODO\((?:b/\d+|[a-z]+)(?:, ?(?:b/\d+|[a-z]+))*\).*\w'
)
_TODO = re.compile(r'\bTODO\b')
# todo-check: enable

# If seen, ignore this line and the next.
_IGNORE = 'todo-check: ignore'

# Ignore a whole section. Please do not change the order of these lines.
_DISABLE = 'todo-check: disable'
_ENABLE = 'todo-check: enable'


def _process_file(ctx: PresubmitContext, todo_pattern: re.Pattern, path: Path):
    with path.open() as ins:
        _LOG.debug('Evaluating path %s', path)
        enabled = True
        prev = ''

        try:
            summary = []
            for i, line in enumerate(ins, 1):
                if _DISABLE in line:
                    enabled = False
                elif _ENABLE in line:
                    enabled = True

                if not enabled or _IGNORE in line or _IGNORE in prev:
                    prev = line
                    continue

                if _TODO.search(line):
                    if not todo_pattern.search(line):
                        # todo-check: ignore
                        ctx.fail(f'Bad TODO on line {i}:', path)
                        ctx.fail(f'    {line.strip()}')
                        summary.append(
                            f'{path.relative_to(ctx.root)}:{i}:{line.strip()}'
                        )

                prev = line

            if summary:
                with ctx.failure_summary_log.open('w') as outs:
                    for line in summary:
                        print(line, file=outs)

        except UnicodeDecodeError:
            # File is not text, like a gif.
            _LOG.debug('File %s is not a text file', path)


def create(
    todo_pattern: re.Pattern = BUGS_ONLY,
    exclude: Iterable[Union[Pattern[str], str]] = EXCLUDE,
):
    """Create a todo_check presubmit step that uses the given pattern."""

    @filter_paths(exclude=exclude)
    def todo_check(ctx: PresubmitContext):
        """Check that TODO lines are valid."""  # todo-check: ignore
        for path in ctx.paths:
            _process_file(ctx, todo_pattern, path)

    return todo_check



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Console for interacting with pw_rpc over HDLC.

This command is no longer supported. Please run pw_system.console instead.

  python -m pw_system.console --device /dev/ttyUSB0 --proto-globs sample.proto
"""

import sys

# TODO(tonymd): Delete this when no longer needed.
PW_RPC_MAX_PACKET_SIZE = 256


def main() -> int:
    print(__doc__)
    return 1


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2019 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Renders HTML documentation using Sphinx."""

# TODO(frolv): Figure out a solution for installing all library dependencies
# to run Sphinx and build RTD docs.

import argparse
import collections
import json
import os
import shutil
import subprocess
import sys

from pathlib import Path
from typing import Dict, List, Optional, Tuple

SCRIPT_HEADER: str = '''
██████╗ ██╗ ██████╗ ██╗    ██╗███████╗███████╗██████╗     ██████╗  ██████╗  ██████╗███████╗
██╔══██╗██║██╔════╝ ██║    ██║██╔════╝██╔════╝██╔══██╗    ██╔══██╗██╔═══██╗██╔════╝██╔════╝
██████╔╝██║██║  ███╗██║ █╗ ██║█████╗  █████╗  ██║  ██║    ██║  ██║██║   ██║██║     ███████╗
██╔═══╝ ██║██║   ██║██║███╗██║██╔══╝  ██╔══╝  ██║  ██║    ██║  ██║██║   ██║██║     ╚════██║
██║     ██║╚██████╔╝╚███╔███╔╝███████╗███████╗██████╔╝    ██████╔╝╚██████╔╝╚██████╗███████║
╚═╝     ╚═╝ ╚═════╝  ╚══╝╚══╝ ╚══════╝╚══════╝╚═════╝     ╚═════╝  ╚═════╝  ╚═════╝╚══════╝
'''


def parse_args() -> argparse.Namespace:
    """Parses command-line arguments."""

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        '--sphinx-build-dir',
        required=True,
        help='Directory in which to build docs',
    )
    parser.add_argument(
        '--conf', required=True, help='Path to conf.py file for Sphinx'
    )
    parser.add_argument(
        '--gn-root', required=True, help='Root of the GN build tree'
    )
    parser.add_argument(
        '--gn-gen-root', required=True, help='Root of the GN gen tree'
    )
    parser.add_argument(
        'sources', nargs='+', help='Paths to the root level rst source files'
    )
    parser.add_argument(
        '--out-dir',
        required=True,
        help='Output directory for rendered HTML docs',
    )
    parser.add_argument(
        '--metadata',
        required=True,
        type=argparse.FileType('r'),
        help='Metadata JSON file',
    )
    parser.add_argument(
        '--google-analytics-id',
        const=None,
        help='Enables Google Analytics with the provided ID',
    )
    return parser.parse_args()


def build_docs(
    src_dir: str, dst_dir: str, google_analytics_id: Optional[str] = None
) -> int:
    """Runs Sphinx to render HTML documentation from a doc tree."""

    # TODO(frolv): Specify the Sphinx script from a prebuilts path instead of
    # requiring it in the tree.
    command = ['sphinx-build', '-W', '-b', 'html', '-d', f'{dst_dir}/help']

    if google_analytics_id is not None:
        command.append(f'-Dgoogle_analytics_id={google_analytics_id}')

    command.extend([src_dir, f'{dst_dir}/html'])

    return subprocess.call(command)


def copy_doc_tree(args: argparse.Namespace) -> None:
    """Copies doc source and input files into a build tree."""

    def build_path(path):
        """Converts a source path to a filename in the build directory."""
        if path.startswith(args.gn_root):
            path = os.path.relpath(path, args.gn_root)
        elif path.startswith(args.gn_gen_root):
            path = os.path.relpath(path, args.gn_gen_root)

        return os.path.join(args.sphinx_build_dir, path)

    source_files = json.load(args.metadata)
    copy_paths = [build_path(f) for f in source_files]

    os.makedirs(args.sphinx_build_dir)
    for source_path in args.sources:
        os.link(
            source_path, f'{args.sphinx_build_dir}/{Path(source_path).name}'
        )
    os.link(args.conf, f'{args.sphinx_build_dir}/conf.py')

    # Map of directory path to list of source and destination file paths.
    dirs: Dict[str, List[Tuple[str, str]]] = collections.defaultdict(list)

    for source_file, copy_path in zip(source_files, copy_paths):
        dirname = os.path.dirname(copy_path)
        dirs[dirname].append((source_file, copy_path))

    for directory, file_pairs in dirs.items():
        os.makedirs(directory, exist_ok=True)
        for src, dst in file_pairs:
            os.link(src, dst)


def main() -> int:
    """Script entry point."""

    args = parse_args()

    # Clear out any existing docs for the target.
    if os.path.exists(args.sphinx_build_dir):
        shutil.rmtree(args.sphinx_build_dir)

    # TODO(b/235349854): Printing the header causes unicode problems on Windows.
    # Disabled for now; re-enable once the root issue is fixed.
    # print(SCRIPT_HEADER)
    copy_doc_tree(args)

    # Flush all script output before running Sphinx.
    print('-' * 80, flush=True)

    return build_docs(
        args.sphinx_build_dir, args.out_dir, args.google_analytics_id
    )


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Install and check status of BoringSSL + Chromium verifier."""

import pathlib
from typing import Sequence
import pw_package.git_repo
import pw_package.package_manager

# List of sources to checkout for chromium verifier.
# The list is hand-picked. It is currently only tested locally (i.e. the list
# compiles and can run certificate chain verification). Unittest will be added
# in pw_tls_client that uses the this package, so that it can be used as a
# criterion for rolling.
CHROMIUM_VERIFIER_LIBRARY_SOURCES = [
    'base/*',
    '!base/check.h',
    '!base/check_op.h',
    '!base/logging.h',
    'build/buildflag.h',
    'build/write_buildflag_header.py',
    'crypto',
    'net/base',
    'net/cert',
    'net/data',
    'net/der',
    'testing/gtest/include',
    'testing/gmock/include',
    'third_party/abseil-cpp',
    'third_party/boringssl',
    'third_party/googletest',
    'time/internal/cctz/include/cctz/civil_time_detail.h',
    'url/gurl.h',
    'url/third_party/mozilla/url_parse.h',
    'url/url_canon.h',
    'url/url_canon_ip.h',
    'url/url_canon_stdstring.h',
    'url/url_constants.h',
    'net/test/test_certificate_data.h',
    'net/cert/internal/path_builder_unittest.cc',
    'third_party/modp_b64',
]

CHROMIUM_VERIFIER_UNITTEST_SOURCES = [
    # TODO(pwbug/394): Look into in necessary unittests to port.
    'net/cert/internal/path_builder_unittest.cc',
]

CHROMIUM_VERIFIER_SOURCES = (
    CHROMIUM_VERIFIER_LIBRARY_SOURCES + CHROMIUM_VERIFIER_UNITTEST_SOURCES
)


def chromium_verifier_repo_path(
    chromium_verifier_install: pathlib.Path,
) -> pathlib.Path:
    """Return the sub-path for repo checkout of chromium verifier"""
    return chromium_verifier_install / 'src'


def chromium_third_party_boringssl_repo_path(
    chromium_verifier_repo: pathlib.Path,
) -> pathlib.Path:
    """Returns the path of third_party/boringssl library in chromium repo"""
    return chromium_verifier_repo / 'third_party' / 'boringssl' / 'src'


def chromium_third_party_googletest_repo_path(
    chromium_verifier_repo: pathlib.Path,
) -> pathlib.Path:
    """Returns the path of third_party/googletest in chromium repo"""
    return chromium_verifier_repo / 'third_party' / 'googletest' / 'src'


class ChromiumVerifier(pw_package.package_manager.Package):
    """Install and check status of Chromium Verifier"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, name='chromium_verifier', **kwargs)
        self._chromium_verifier = pw_package.git_repo.GitRepo(
            name='chromium_verifier',
            url='https://chromium.googlesource.com/chromium/src',
            commit='04ebce24d98339954fb1d2a67e68da7ca81ca47c',
            sparse_list=CHROMIUM_VERIFIER_SOURCES,
        )

        # The following is for checking out necessary headers of
        # boringssl and googletest third party libraries that chromium verifier
        # depends on. The actual complete libraries will be separate packages.

        self._boringssl = pw_package.git_repo.GitRepo(
            name='boringssl',
            url=''.join(
                [
                    'https://pigweed.googlesource.com',
                    '/third_party/boringssl/boringssl',
                ]
            ),
            commit='9f55d972854d0b34dae39c7cd3679d6ada3dfd5b',
            sparse_list=['include'],
        )

        self._googletest = pw_package.git_repo.GitRepo(
            name='googletest',
            url=''.join(
                [
                    'https://chromium.googlesource.com/',
                    'external/github.com/google/googletest.git',
                ]
            ),
            commit='53495a2a7d6ba7e0691a7f3602e9a5324bba6e45',
            sparse_list=[
                'googletest/include',
                'googlemock/include',
            ],
        )

    def install(self, path: pathlib.Path) -> None:
        # Checkout chromium verifier
        chromium_repo = chromium_verifier_repo_path(path)
        self._chromium_verifier.install(chromium_repo)

        # Checkout third party boringssl headers
        boringssl_repo = chromium_third_party_boringssl_repo_path(chromium_repo)
        self._boringssl.install(boringssl_repo)

        # Checkout third party googletest headers
        googletest_repo = chromium_third_party_googletest_repo_path(
            chromium_repo
        )
        self._googletest.install(googletest_repo)

    def status(self, path: pathlib.Path) -> bool:
        chromium_repo = chromium_verifier_repo_path(path)
        if not self._chromium_verifier.status(chromium_repo):
            return False

        boringssl_repo = chromium_third_party_boringssl_repo_path(chromium_repo)
        if not self._boringssl.status(boringssl_repo):
            return False

        googletest_repo = chromium_third_party_googletest_repo_path(
            chromium_repo
        )
        if not self._googletest.status(googletest_repo):
            return False

        return True

    def info(self, path: pathlib.Path) -> Sequence[str]:
        return (
            f'{self.name} installed in: {path}',
            'Enable by running "gn args out" and adding this line:',
            f'  dir_pw_third_party_chromium_verifier = {path}',
        )


pw_package.package_manager.register(ChromiumVerifier)



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Provides general purpose plugin functionality.

As used in this module, a plugin is a Python object associated with a name.
Plugins are registered in a Registry. The plugin object is typically a function,
but can be anything.

Plugins may be loaded in a variety of ways:

- Listed in a plugins file in the file system (e.g. as "name module target").
- Registered in a Python file using a decorator (@my_registry.plugin).
- Registered directly or by name with function calls on a registry object.

This functionality can be used to create plugins for command line tools,
interactive consoles, or anything else. Pigweed's pw command uses this module
for its plugins.
"""

import collections
import collections.abc
import importlib
import inspect
import logging
from pathlib import Path
import pkgutil
import sys
from textwrap import TextWrapper
import types
from typing import Any, Callable, Dict, List, Iterable, Iterator, Optional, Set

_LOG = logging.getLogger(__name__)
_BUILT_IN = '<built-in>'


class Error(Exception):
    """Indicates that a plugin is invalid or cannot be registered."""

    def __str__(self):
        """Displays the error as a string, including the __cause__ if present.

        Adding __cause__ gives useful context without displaying a backtrace.
        """
        if self.__cause__ is None:
            return super().__str__()

        return (
            f'{super().__str__()} '
            f'({type(self.__cause__).__name__}: {self.__cause__})'
        )


def _get_module(member: object) -> types.ModuleType:
    """Gets the module or a fake module if the module isn't found."""
    module = inspect.getmodule(member)
    return module if module else types.ModuleType('<unknown>')


class Plugin:
    """Represents a Python entity registered as a plugin.

    Each plugin resolves to a Python object, typically a function.
    """

    @classmethod
    def from_name(
        cls,
        name: str,
        module_name: str,
        member_name: str,
        source: Optional[Path],
    ) -> 'Plugin':
        """Creates a plugin by module and attribute name.

        Args:
          name: the name of the plugin
          module_name: Python module name (e.g. 'foo_pkg.bar')
          member_name: the name of the member in the module
          source: path to the plugins file that declared this plugin, if any
        """

        # Attempt to access the module and member. Catch any errors that might
        # occur, since a bad plugin shouldn't be a fatal error.
        try:
            module = importlib.import_module(module_name)
        except Exception as err:
            _LOG.debug(
                'Failed to import module "%s" for "%s" plugin',
                module_name,
                name,
                exc_info=True,
            )
            raise Error(f'Failed to import module "{module_name}"') from err

        try:
            member = getattr(module, member_name)
        except AttributeError as err:
            raise Error(
                f'"{module_name}.{member_name}" does not exist'
            ) from err

        return cls(name, member, source)

    def __init__(
        self, name: str, target: Any, source: Optional[Path] = None
    ) -> None:
        """Creates a plugin for the provided target."""
        self.name = name
        self._module = _get_module(target)
        self.target = target
        self.source = source

    @property
    def target_name(self) -> str:
        return (
            f'{self._module.__name__}.'
            f'{getattr(self.target, "__name__", self.target)}'
        )

    @property
    def source_name(self) -> str:
        return _BUILT_IN if self.source is None else str(self.source)

    def run_with_argv(self, argv: Iterable[str]) -> int:
        """Sets sys.argv and calls the plugin function.

        This is used to call a plugin as if from the command line.
        """
        original_sys_argv = sys.argv
        sys.argv = [f'pw {self.name}', *argv]

        try:
            return self.target()
        finally:
            sys.argv = original_sys_argv

    def help(self, full: bool = False) -> str:
        """Returns a description of this plugin from its docstring."""
        docstring = self.target.__doc__ or self._module.__doc__ or ''
        return docstring if full else next(iter(docstring.splitlines()), '')

    def details(self, full: bool = False) -> Iterator[str]:
        yield f'help    {self.help(full=full)}'
        yield f'module  {self._module.__name__}'
        yield f'target  {getattr(self.target, "__name__", self.target)}'
        yield f'source  {self.source_name}'

    def __repr__(self) -> str:
        return (
            f'{self.__class__.__name__}(name={self.name!r}, '
            f'target={self.target_name}'
            f'{f", source={self.source_name!r}" if self.source else ""})'
        )


def callable_with_no_args(plugin: Plugin) -> None:
    """Checks that a plugin is callable without arguments.

    May be used for the validator argument to Registry.
    """
    try:
        params = inspect.signature(plugin.target).parameters
    except TypeError:
        raise Error(
            'Plugin functions must be callable, but '
            f'{plugin.target_name} is a '
            f'{type(plugin.target).__name__}'
        )

    positional = sum(p.default == p.empty for p in params.values())
    if positional:
        raise Error(
            f'Plugin functions cannot have any required positional '
            f'arguments, but {plugin.target_name} has {positional}'
        )


class Registry(collections.abc.Mapping):
    """Manages a set of plugins from Python modules or plugins files."""

    def __init__(
        self, validator: Callable[[Plugin], Any] = lambda _: None
    ) -> None:
        """Creates a new, empty plugins registry.

        Args:
          validator: Function that checks whether a plugin is valid and should
              be registered. Must raise plugins.Error is the plugin is invalid.
        """

        self._registry: Dict[str, Plugin] = {}
        self._sources: Set[Path] = set()  # Paths to plugins files
        self._errors: Dict[str, List[Exception]] = collections.defaultdict(list)
        self._validate_plugin = validator

    def __getitem__(self, name: str) -> Plugin:
        """Accesses a plugin by name; raises KeyError if it does not exist."""
        if name in self._registry:
            return self._registry[name]

        if name in self._errors:
            raise KeyError(
                f'Registration for "{name}" failed: '
                + ', '.join(str(e) for e in self._errors[name])
            )

        raise KeyError(f'The plugin "{name}" has not been registered')

    def __iter__(self) -> Iterator[str]:
        return iter(self._registry)

    def __len__(self) -> int:
        return len(self._registry)

    def errors(self) -> Dict[str, List[Exception]]:
        return self._errors

    def run_with_argv(self, name: str, argv: Iterable[str]) -> int:
        """Runs a plugin by name, setting sys.argv to the provided args.

        This is used to run a command as if it were executed directly from the
        command line. The plugin is expected to return an int.

        Raises:
          KeyError if plugin is not registered.
        """
        return self[name].run_with_argv(argv)

    def _should_register(self, plugin: Plugin) -> bool:
        """Determines and logs if a plugin should be registered or not.

        Some errors are exceptions, others are not.
        """

        if plugin.name in self._registry and plugin.source is None:
            raise Error(
                f'Attempted to register built-in plugin "{plugin.name}", but '
                'a plugin with that name was previously registered '
                f'({self[plugin.name]})!'
            )

        # Run the user-provided validation function, which raises exceptions
        # if there are errors.
        self._validate_plugin(plugin)

        existing = self._registry.get(plugin.name)

        if existing is None:
            return True

        if existing.source is None:
            _LOG.debug(
                '%s: Overriding built-in plugin "%s" with %s',
                plugin.source_name,
                plugin.name,
                plugin.target_name,
            )
            return True

        if plugin.source != existing.source:
            _LOG.debug(
                '%s: The plugin "%s" was previously registered in %s; '
                'ignoring registration as %s',
                plugin.source_name,
                plugin.name,
                self._registry[plugin.name].source,
                plugin.target_name,
            )
        elif plugin.source not in self._sources:
            _LOG.warning(
                '%s: "%s" is registered file multiple times in this file! '
                'Only the first registration takes effect',
                plugin.source_name,
                plugin.name,
            )

        return False

    def register(self, name: str, target: Any) -> Optional[Plugin]:
        """Registers an object as a plugin."""
        return self._register(Plugin(name, target, None))

    def register_by_name(
        self,
        name: str,
        module_name: str,
        member_name: str,
        source: Optional[Path] = None,
    ) -> Optional[Plugin]:
        """Registers an object from its module and name as a plugin."""
        return self._register(
            Plugin.from_name(name, module_name, member_name, source)
        )

    def _register(self, plugin: Plugin) -> Optional[Plugin]:
        # Prohibit functions not from a plugins file from overriding others.
        if not self._should_register(plugin):
            return None

        self._registry[plugin.name] = plugin
        _LOG.debug(
            '%s: Registered plugin "%s" for %s',
            plugin.source_name,
            plugin.name,
            plugin.target_name,
        )

        return plugin

    def register_file(self, path: Path) -> None:
        """Registers plugins from a plugins file.

        Any exceptions raised from parsing the file are caught and logged.
        """
        with path.open() as contents:
            for lineno, line in enumerate(contents, 1):
                line = line.strip()
                if not line or line.startswith('#'):
                    continue

                try:
                    name, module, function = line.split()
                except ValueError as err:
                    self._errors[line.strip()].append(Error(err))
                    _LOG.error(
                        '%s:%d: Failed to parse plugin entry "%s": '
                        'Expected 3 items (name, module, function), '
                        'got %d',
                        path,
                        lineno,
                        line,
                        len(line.split()),
                    )
                    continue

                try:
                    self.register_by_name(name, module, function, path)
                except Error as err:
                    self._errors[name].append(err)
                    _LOG.error(
                        '%s: Failed to register plugin "%s": %s',
                        path,
                        name,
                        err,
                    )

        self._sources.add(path)

    def register_directory(
        self,
        directory: Path,
        file_name: str,
        restrict_to: Optional[Path] = None,
    ) -> None:
        """Finds and registers plugins from plugins files in a directory.

        Args:
          directory: The directory from which to start searching up.
          file_name: The name of plugins files to look for.
          restrict_to: If provided, do not search higher than this directory.
        """
        for path in find_all_in_parents(file_name, directory):
            if not path.is_file():
                continue

            if restrict_to is not None and restrict_to not in path.parents:
                _LOG.debug(
                    "Skipping plugins file %s because it's outside of %s",
                    path,
                    restrict_to,
                )
                continue

            _LOG.debug('Found plugins file %s', path)
            self.register_file(path)

    def short_help(self) -> str:
        """Returns a help string for the registered plugins."""
        width = (
            max(len(name) for name in self._registry) + 1
            if self._registry
            else 1
        )
        help_items = '\n'.join(
            f'  {name:{width}} {plugin.help()}'
            for name, plugin in sorted(self._registry.items())
        )
        return f'supported plugins:\n{help_items}'

    def detailed_help(self, plugins: Iterable[str] = ()) -> Iterator[str]:
        """Yields lines of detailed information about commands."""
        if not plugins:
            plugins = list(self._registry)

        yield '\ndetailed plugin information:'

        wrapper = TextWrapper(
            width=80, initial_indent='   ', subsequent_indent=' ' * 11
        )

        plugins = sorted(plugins)
        for plugin in plugins:
            yield f'  [{plugin}]'

            try:
                for line in self[plugin].details(full=len(plugins) == 1):
                    yield wrapper.fill(line)
            except KeyError as err:
                yield wrapper.fill(f'error   {str(err)[1:-1]}')

            yield ''

        yield 'Plugins files:'

        if self._sources:
            yield from (
                f'  [{i}] {file}' for i, file in enumerate(self._sources, 1)
            )
        else:
            yield '  (none found)'

    def plugin(
        self, function: Optional[Callable] = None, *, name: Optional[str] = None
    ) -> Callable[[Callable], Callable]:
        """Decorator that registers a function with this plugin registry."""

        def decorator(function: Callable) -> Callable:
            self.register(function.__name__ if name is None else name, function)
            return function

        if function is None:
            return decorator

        self.register(function.__name__, function)
        return function


def find_in_parents(name: str, path: Path) -> Optional[Path]:
    """Searches parent directories of the path for a file or directory."""
    path = path.resolve()

    while not path.joinpath(name).exists():
        path = path.parent

        if path.samefile(path.parent):
            return None

    return path.joinpath(name)


def find_all_in_parents(name: str, path: Path) -> Iterator[Path]:
    """Searches all parent directories of the path for files or directories."""

    while True:
        result = find_in_parents(name, path)
        if result is None:
            return

        yield result
        path = result.parent.parent


def import_submodules(
    module: types.ModuleType, recursive: bool = False
) -> None:
    """Imports the submodules of a package.

    This can be used to collect plugins registered with a decorator from a
    directory.
    """
    path = module.__path__  # type: ignore[attr-defined]
    if recursive:
        modules = pkgutil.walk_packages(path, module.__name__ + '.')
    else:
        modules = pkgutil.iter_modules(path, module.__name__ + '.')

    for info in modules:
        importlib.import_module(info.name)



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""pw_ide exceptions."""


class UnsupportedPlatformException(Exception):
    """Raised when an action is attempted on an unsupported platform."""


class InvalidTargetException(Exception):
    """Exception for invalid compilation targets."""


class BadCompDbException(Exception):
    """Exception for compliation databases that don't conform to the format."""


class MissingCompDbException(Exception):
    """Exception for missing compilation database files."""


class UnresolvablePathException(Exception):
    """Raised when an ambiguous path cannot be resolved."""



#!/usr/bin/env python
# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Watch build config dataclasses."""

from dataclasses import dataclass, field
import logging
from pathlib import Path
import shlex
from typing import Callable, Dict, List, Optional, TYPE_CHECKING

from prompt_toolkit.formatted_text import ANSI, StyleAndTextTuples
from prompt_toolkit.formatted_text.base import OneStyleAndTextTuple

if TYPE_CHECKING:
    from pw_build.project_builder import ProjectBuilder
    from pw_build.project_builder_prefs import ProjectBuilderPrefs

_LOG = logging.getLogger('pw_build.watch')


class UnknownBuildSystem(Exception):
    """Exception for requesting unsupported build systems."""


class UnknownBuildDir(Exception):
    """Exception for an unknown build dir before command running."""


@dataclass
class BuildCommand:
    """Store details of a single build step.

    Example usage:

    .. code-block:: python

        from pw_build.build_recipe import BuildCommand, BuildRecipe

        def should_gen_gn(out: Path):
            return not (out / 'build.ninja').is_file()

        cmd1 = BuildCommand(build_dir='out',
                            command=['gn', 'gen', '{build_dir}'],
                            run_if=should_gen_gn)

        cmd2 = BuildCommand(build_dir='out',
                            build_system_command='ninja',
                            build_system_extra_args=['-k', '0'],
                            targets=['default']),

    Args:
        build_dir: Output directory for this build command. This can be omitted
            if the BuildCommand is included in the steps of a BuildRecipe.
        build_system_command: This command should end with ``ninja``, ``make``,
            or ``bazel``.
        build_system_extra_args: A list of extra arguments passed to the
            build_system_command. If running ``bazel test`` include ``test`` as
            an extra arg here.
        targets: Optional list of targets to build in the build_dir.
        command: List of strings to run as a command. These are passed to
            subprocess.run(). Any instances of the ``'{build_dir}'`` string
            literal will be replaced at run time with the out directory.
        run_if: A callable function to run before executing this
            BuildCommand. The callable takes one Path arg for the build_dir. If
            the callable returns true this command is executed. All
            BuildCommands are run by default.
    """

    build_dir: Optional[Path] = None
    build_system_command: Optional[str] = None
    build_system_extra_args: List[str] = field(default_factory=list)
    targets: List[str] = field(default_factory=list)
    command: List[str] = field(default_factory=list)
    run_if: Callable[[Path], bool] = lambda _build_dir: True

    def __post_init__(self) -> None:
        # Copy self._expanded_args from the command list.
        self._expanded_args: List[str] = []
        if self.command:
            self._expanded_args = self.command

    def should_run(self) -> bool:
        """Return True if this build command should be run."""
        if self.build_dir:
            return self.run_if(self.build_dir)
        return True

    def _get_starting_build_system_args(self) -> List[str]:
        """Return flags that appear immediately after the build command."""
        assert self.build_system_command
        assert self.build_dir
        if self.build_system_command.endswith('bazel'):
            return ['--output_base', str(self.build_dir)]
        return []

    def _get_build_system_args(self) -> List[str]:
        assert self.build_system_command
        assert self.build_dir

        # Both make and ninja use -C for a build directory.
        if self.build_system_command.endswith(
            'make'
        ) or self.build_system_command.endswith('ninja'):
            return ['-C', str(self.build_dir), *self.targets]

        # Bazel relies on --output_base which is handled by the
        # _get_starting_build_system_args() function.
        if self.build_system_command.endswith('bazel'):
            return [*self.targets]

        raise UnknownBuildSystem(
            f'\n\nUnknown build system command "{self.build_system_command}" '
            f'for build directory "{self.build_dir}".\n'
            'Supported commands: ninja, bazel, make'
        )

    def _resolve_expanded_args(self) -> List[str]:
        """Replace instances of '{build_dir}' with the self.build_dir."""
        resolved_args = []
        for arg in self._expanded_args:
            if arg == "{build_dir}":
                if not self.build_dir:
                    raise UnknownBuildDir(
                        '\n\nUnknown "{build_dir}" value for command:\n'
                        f'  {self._expanded_args}\n'
                        f'In BuildCommand: {repr(self)}\n\n'
                        'Check build_dir is set for the above BuildCommand'
                        'or included as a step to a BuildRecipe.'
                    )
                resolved_args.append(str(self.build_dir.resolve()))
            else:
                resolved_args.append(arg)
        return resolved_args

    def ninja_command(self) -> bool:
        if self.build_system_command and self.build_system_command.endswith(
            'ninja'
        ):
            return True
        return False

    def bazel_command(self) -> bool:
        if self.build_system_command and self.build_system_command.endswith(
            'bazel'
        ):
            return True
        return False

    def bazel_build_command(self) -> bool:
        if self.bazel_command() and 'build' in self.build_system_extra_args:
            return True
        return False

    def bazel_clean_command(self) -> bool:
        if self.bazel_command() and 'clean' in self.build_system_extra_args:
            return True
        return False

    def get_args(
        self,
        additional_ninja_args: Optional[List[str]] = None,
        additional_bazel_args: Optional[List[str]] = None,
        additional_bazel_build_args: Optional[List[str]] = None,
    ) -> List[str]:
        """Return all args required to launch this BuildCommand."""
        # If this is a plain command step, return self._expanded_args as-is.
        if not self.build_system_command:
            return self._resolve_expanded_args()

        # Assmemble user-defined extra args.
        extra_args = []
        extra_args.extend(self.build_system_extra_args)
        if additional_ninja_args and self.ninja_command():
            extra_args.extend(additional_ninja_args)

        if additional_bazel_build_args and self.bazel_build_command():
            extra_args.extend(additional_bazel_build_args)

        if additional_bazel_args and self.bazel_command():
            extra_args.extend(additional_bazel_args)

        build_system_target_args = []
        if not self.bazel_clean_command():
            build_system_target_args = self._get_build_system_args()

        # Construct the build system command args.
        command = [
            self.build_system_command,
            *self._get_starting_build_system_args(),
            *extra_args,
            *build_system_target_args,
        ]
        return command

    def __str__(self) -> str:
        return ' '.join(shlex.quote(arg) for arg in self.get_args())


@dataclass
class BuildRecipeStatus:
    """Stores the status of a build recipe."""

    recipe: 'BuildRecipe'
    current_step: str = ''
    percent: float = 0.0
    error_count: int = 0
    return_code: Optional[int] = None
    flag_done: bool = False
    flag_started: bool = False
    error_lines: Dict[int, List[str]] = field(default_factory=dict)

    def pending(self) -> bool:
        return self.return_code is None

    def failed(self) -> bool:
        if self.return_code is not None:
            return self.return_code != 0
        return False

    def append_failure_line(self, line: str) -> None:
        lines = self.error_lines.get(self.error_count, [])
        lines.append(line)
        self.error_lines[self.error_count] = lines

    def has_empty_ninja_errors(self) -> bool:
        for error_lines in self.error_lines.values():
            # NOTE: There will be at least 2 lines for each ninja failure:
            # - A starting 'FAILED: target' line
            # - An ending line with this format:
            #   'ninja: error: ... cannot make progress due to previous errors'

            # If the total error line count is very short, assume it's an empty
            # ninja error.
            if len(error_lines) <= 3:
                # If there is a failure in the regen step, there will be 3 error
                # lines: The above two and one more with the regen command.
                return True
            # Otherwise, if the line starts with FAILED: build.ninja the failure
            # is likely in the regen step and there will be extra cmake or gn
            # error text that was not captured.
            for line in error_lines:
                if line.startswith(
                    '\033[31mFAILED: \033[0mbuild.ninja'
                ) or line.startswith('FAILED: build.ninja'):
                    return True
        return False

    def increment_error_count(self, count: int = 1) -> None:
        self.error_count += count
        if self.error_count not in self.error_lines:
            self.error_lines[self.error_count] = []

    def should_log_failures(self) -> bool:
        return (
            self.recipe.project_builder is not None
            and self.recipe.project_builder.separate_build_file_logging
            and (not self.recipe.project_builder.send_recipe_logs_to_root)
        )

    def log_last_failure(self) -> None:
        """Log the last ninja error if available."""
        if not self.should_log_failures():
            return

        logger = self.recipe.error_logger
        if not logger:
            return

        _color = self.recipe.project_builder.color  # type: ignore

        lines = self.error_lines.get(self.error_count, [])
        _LOG.error('')
        _LOG.error(' ╔════════════════════════════════════')
        _LOG.error(
            ' ║  START %s Failure #%d:',
            _color.cyan(self.recipe.display_name),
            self.error_count,
        )

        logger.error('')
        for line in lines:
            logger.error(line)
        logger.error('')

        _LOG.error(
            ' ║  END %s Failure #%d',
            _color.cyan(self.recipe.display_name),
            self.error_count,
        )
        _LOG.error(" ╚════════════════════════════════════")
        _LOG.error('')

    def log_entire_recipe_logfile(self) -> None:
        """Log the entire build logfile if no ninja errors available."""
        if not self.should_log_failures():
            return

        recipe_logfile = self.recipe.logfile
        if not recipe_logfile:
            return

        _color = self.recipe.project_builder.color  # type: ignore

        logfile_path = str(recipe_logfile.resolve())

        _LOG.error('')
        _LOG.error(' ╔════════════════════════════════════')
        _LOG.error(
            ' ║  %s Failure; Entire log below:',
            _color.cyan(self.recipe.display_name),
        )
        _LOG.error(' ║  %s %s', _color.yellow('START'), logfile_path)

        logger = self.recipe.error_logger
        if not logger:
            return

        logger.error('')
        for line in recipe_logfile.read_text(
            encoding='utf-8', errors='ignore'
        ).splitlines():
            logger.error(line)
        logger.error('')

        _LOG.error(' ║  %s %s', _color.yellow('END'), logfile_path)
        _LOG.error(" ╚════════════════════════════════════")
        _LOG.error('')

    def status_slug(self, restarting: bool = False) -> OneStyleAndTextTuple:
        status = ('', '')
        if not self.recipe.enabled:
            return ('fg:ansidarkgray', 'Disabled')

        waiting = False
        if self.done:
            if self.passed():
                status = ('fg:ansigreen', 'OK      ')
            elif self.failed():
                status = ('fg:ansired', 'FAIL    ')
        elif self.started:
            status = ('fg:ansiyellow', 'Building')
        else:
            waiting = True
            status = ('default', 'Waiting ')

        # Only show Aborting if the process is building (or has failures).
        if restarting and not waiting and not self.passed():
            status = ('fg:ansiyellow', 'Aborting')
        return status

    def current_step_formatted(self) -> StyleAndTextTuples:
        formatted_text: StyleAndTextTuples = []
        if self.passed():
            return formatted_text

        if self.current_step:
            if '\x1b' in self.current_step:
                formatted_text = ANSI(self.current_step).__pt_formatted_text__()
            else:
                formatted_text = [('', self.current_step)]

        return formatted_text

    @property
    def done(self) -> bool:
        return self.flag_done

    @property
    def started(self) -> bool:
        return self.flag_started

    def mark_done(self) -> None:
        self.flag_done = True

    def mark_started(self) -> None:
        self.flag_started = True

    def set_failed(self) -> None:
        self.flag_done = True
        self.return_code = -1

    def set_passed(self) -> None:
        self.flag_done = True
        self.return_code = 0

    def passed(self) -> bool:
        if self.done and self.return_code is not None:
            return self.return_code == 0
        return False


@dataclass
class BuildRecipe:
    """Dataclass to store a list of BuildCommands.

    Example usage:

    .. code-block:: python

        from pw_build.build_recipe import BuildCommand, BuildRecipe

        def should_gen_gn(out: Path) -> bool:
            return not (out / 'build.ninja').is_file()

        recipe = BuildRecipe(
            build_dir='out',
            title='Vanilla Ninja Build',
            steps=[
                BuildCommand(command=['gn', 'gen', '{build_dir}'],
                             run_if=should_gen_gn),
                BuildCommand(build_system_command='ninja',
                             build_system_extra_args=['-k', '0'],
                             targets=['default']),
            ],
        )

    Args:
        build_dir: Output directory for this BuildRecipe. On init this out dir
            is set for all included steps.
        steps: List of BuildCommands to run.
        title: Custom title. The build_dir is used if this is ommited.
    """

    build_dir: Path
    steps: List[BuildCommand] = field(default_factory=list)
    title: Optional[str] = None
    enabled: bool = True

    def __hash__(self):
        return hash((self.build_dir, self.title, len(self.steps)))

    def __post_init__(self) -> None:
        # Update all included steps to use this recipe's build_dir.
        for step in self.steps:
            if self.build_dir:
                step.build_dir = self.build_dir

        # Set logging variables
        self._logger: Optional[logging.Logger] = None
        self.error_logger: Optional[logging.Logger] = None
        self._logfile: Optional[Path] = None
        self._status: BuildRecipeStatus = BuildRecipeStatus(self)
        self.project_builder: Optional['ProjectBuilder'] = None

    def toggle_enabled(self) -> None:
        self.enabled = not self.enabled

    def set_project_builder(self, project_builder) -> None:
        self.project_builder = project_builder

    def set_targets(self, new_targets: List[str]) -> None:
        """Reset all build step targets."""
        for step in self.steps:
            step.targets = new_targets

    def set_logger(self, logger: logging.Logger) -> None:
        self._logger = logger

    def set_error_logger(self, logger: logging.Logger) -> None:
        self.error_logger = logger

    def set_logfile(self, log_file: Path) -> None:
        self._logfile = log_file

    def reset_status(self) -> None:
        self._status = BuildRecipeStatus(self)

    @property
    def status(self) -> BuildRecipeStatus:
        return self._status

    @property
    def log(self) -> logging.Logger:
        if self._logger:
            return self._logger
        return logging.getLogger()

    @property
    def logfile(self) -> Optional[Path]:
        return self._logfile

    @property
    def display_name(self) -> str:
        if self.title:
            return self.title
        return str(self.build_dir)

    def targets(self) -> List[str]:
        return list(
            set(target for step in self.steps for target in step.targets)
        )

    def __str__(self) -> str:
        message = self.display_name
        targets = self.targets()
        if targets:
            target_list = ' '.join(self.targets())
            message = f'{message} -- {target_list}'
        return message


def create_build_recipes(prefs: 'ProjectBuilderPrefs') -> List[BuildRecipe]:
    """Create a list of BuildRecipes from ProjectBuilderPrefs."""
    build_recipes: List[BuildRecipe] = []

    if prefs.run_commands:
        for command_str in prefs.run_commands:
            build_recipes.append(
                BuildRecipe(
                    build_dir=Path.cwd(),
                    steps=[BuildCommand(command=shlex.split(command_str))],
                    title=command_str,
                )
            )

    for build_dir, targets in prefs.build_directories.items():
        steps: List[BuildCommand] = []
        build_path = Path(build_dir)
        if not targets:
            targets = []

        for (
            build_system_command,
            build_system_extra_args,
        ) in prefs.build_system_commands(build_dir):
            steps.append(
                BuildCommand(
                    build_system_command=build_system_command,
                    build_system_extra_args=build_system_extra_args,
                    targets=targets,
                )
            )

        build_recipes.append(
            BuildRecipe(
                build_dir=build_path,
                steps=steps,
            )
        )

    return build_recipes



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""This module manages the global plugin registry for pw_cli."""

import argparse
import os
from pathlib import Path
import sys
from typing import Iterable

from pw_cli import arguments, plugins

_plugin_registry = plugins.Registry(validator=plugins.callable_with_no_args)
REGISTRY_FILE = 'PW_PLUGINS'


def _register_builtin_plugins(registry: plugins.Registry) -> None:
    """Registers the commands that are included with pw by default."""

    # Register these by name to avoid circular dependencies.
    registry.register_by_name('bloat', 'pw_bloat.__main__', 'main')
    registry.register_by_name('doctor', 'pw_doctor.doctor', 'main')
    registry.register_by_name('format', 'pw_presubmit.format_code', 'main')
    registry.register_by_name('keep-sorted', 'pw_presubmit.keep_sorted', 'main')
    registry.register_by_name('logdemo', 'pw_cli.log', 'main')
    registry.register_by_name('module', 'pw_module.__main__', 'main')
    registry.register_by_name(
        'python-packages', 'pw_env_setup.python_packages', 'main'
    )
    registry.register_by_name('test', 'pw_unit_test.test_runner', 'main')
    registry.register_by_name('watch', 'pw_watch.watch', 'main')

    registry.register('help', _help_command)


def _help_command():
    """Display detailed information about pw commands."""
    parser = argparse.ArgumentParser(description=_help_command.__doc__)
    parser.add_argument(
        'plugins',
        metavar='plugin',
        nargs='*',
        help='command for which to display detailed info',
    )

    print(arguments.format_help(_plugin_registry), file=sys.stderr)

    for line in _plugin_registry.detailed_help(**vars(parser.parse_args())):
        print(line, file=sys.stderr)


def register(directory: Path) -> None:
    _register_builtin_plugins(_plugin_registry)
    _plugin_registry.register_directory(
        directory, REGISTRY_FILE, Path(os.environ.get('PW_PROJECT_ROOT', ''))
    )


def errors() -> dict:
    return _plugin_registry.errors()


def format_help() -> str:
    return arguments.format_help(_plugin_registry)


def run(name: str, args: Iterable[str]) -> int:
    return _plugin_registry.run_with_argv(name, args)



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""pw_console embed class."""

import asyncio
import logging
from pathlib import Path
from typing import Any, Dict, List, Iterable, Optional, Tuple, Union

from prompt_toolkit.completion import WordCompleter

from pw_console.console_app import ConsoleApp
from pw_console.get_pw_console_app import PW_CONSOLE_APP_CONTEXTVAR
from pw_console.plugin_mixin import PluginMixin
from pw_console.python_logging import (
    setup_python_logging as pw_console_setup_python_logging,
)
from pw_console.widgets import (
    FloatingWindowPane,
    WindowPane,
    WindowPaneToolbar,
)


def _set_console_app_instance(plugin: Any, console_app: ConsoleApp) -> None:
    if hasattr(plugin, 'pw_console_init'):
        plugin.pw_console_init(console_app)
    else:
        plugin.application = console_app


class PwConsoleEmbed:
    """Embed class for customizing the console before startup."""

    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        global_vars=None,
        local_vars=None,
        loggers: Optional[
            Union[Dict[str, Iterable[logging.Logger]], Iterable]
        ] = None,
        test_mode=False,
        repl_startup_message: Optional[str] = None,
        help_text: Optional[str] = None,
        app_title: Optional[str] = None,
        config_file_path: Optional[Union[str, Path]] = None,
    ) -> None:
        """Call this to embed pw console at the call point within your program.

        Example usage:

        .. code-block:: python

            import logging

            from pw_console import PwConsoleEmbed

            # Create the pw_console embed instance
            console = PwConsoleEmbed(
                global_vars=globals(),
                local_vars=locals(),
                loggers={
                    'Host Logs': [
                        logging.getLogger(__package__),
                        logging.getLogger(__name__),
                    ],
                    'Device Logs': [
                        logging.getLogger('usb_gadget'),
                    ],
                },
                app_title='My Awesome Console',
                config_file_path='/home/user/project/.pw_console.yaml',
            )
            # Optional: Add custom completions
            console.add_sentence_completer(
                {
                    'some_function', 'Function',
                    'some_variable', 'Variable',
                }
            )

            # Setup Python loggers to output to a file instead of STDOUT.
            console.setup_python_logging()

            # Then run the console with:
            console.embed()

        Args:
            global_vars: Dictionary representing the desired global symbol
                table. Similar to what is returned by `globals()`.
            local_vars: Dictionary representing the desired local symbol
                table. Similar to what is returned by `locals()`.
            loggers: Dict with keys of log window titles and values of either:

                    1. List of `logging.getLogger()
                       <https://docs.python.org/3/library/logging.html#logging.getLogger>`_
                       instances.
                    2. A single pw_console.log_store.LogStore instance.

            app_title: Custom title text displayed in the user interface.
            repl_startup_message: Custom text shown by default in the repl
                output pane.
            help_text: Custom text shown at the top of the help window before
                keyboard shortcuts.
            config_file_path: Path to a pw_console yaml config file.
        """

        self.global_vars = global_vars
        self.local_vars = local_vars
        self.loggers = loggers
        self.test_mode = test_mode
        self.repl_startup_message = repl_startup_message
        self.help_text = help_text
        self.app_title = app_title
        self.config_file_path = (
            Path(config_file_path) if config_file_path else None
        )

        self.console_app: Optional[ConsoleApp] = None
        self.extra_completers: List = []

        self.setup_python_logging_called = False
        self.hidden_by_default_windows: List[str] = []
        self.window_plugins: List[WindowPane] = []
        self.floating_window_plugins: List[Tuple[FloatingWindowPane, Dict]] = []
        self.top_toolbar_plugins: List[WindowPaneToolbar] = []
        self.bottom_toolbar_plugins: List[WindowPaneToolbar] = []

    def add_window_plugin(self, window_pane: WindowPane) -> None:
        """Include a custom window pane plugin.

        Args:
            window_pane: Any instance of the WindowPane class.
        """
        self.window_plugins.append(window_pane)

    def add_floating_window_plugin(
        self, window_pane: FloatingWindowPane, **float_args
    ) -> None:
        """Include a custom floating window pane plugin.

        This adds a FloatingWindowPane class to the pw_console UI. The first
        argument should be the window to add and the remaining keyword arguments
        are passed to the prompt_toolkit Float() class. This allows positioning
        of the floating window. By default the floating window will be
        centered. To anchor the window to a side or corner of the screen set the
        ``left``, ``right``, ``top``, or ``bottom`` keyword args.

        For example:

        .. code-block:: python

           from pw_console import PwConsoleEmbed

           console = PwConsoleEmbed(...)
           my_plugin = MyPlugin()
           # Anchor this floating window 2 rows away from the top and 4 columns
           # away from the left edge of the screen.
           console.add_floating_window_plugin(my_plugin, top=2, left=4)

        See all possible keyword args in the prompt_toolkit documentation:
        https://python-prompt-toolkit.readthedocs.io/en/stable/pages/reference.html#prompt_toolkit.layout.Float

        Args:
            window_pane: Any instance of the FloatingWindowPane class.
            left: Distance to the left edge of the screen
            right: Distance to the right edge of the screen
            top: Distance to the top edge of the screen
            bottom: Distance to the bottom edge of the screen
        """
        self.floating_window_plugins.append((window_pane, float_args))

    def add_top_toolbar(self, toolbar: WindowPaneToolbar) -> None:
        """Include a toolbar plugin to display on the top of the screen.

        Top toolbars appear above all window panes and just below the main menu
        bar. They span the full width of the screen.

        Args:
            toolbar: Instance of the WindowPaneToolbar class.
        """
        self.top_toolbar_plugins.append(toolbar)

    def add_bottom_toolbar(self, toolbar: WindowPaneToolbar) -> None:
        """Include a toolbar plugin to display at the bottom of the screen.

        Bottom toolbars appear below all window panes and span the full width of
        the screen.

        Args:
            toolbar: Instance of the WindowPaneToolbar class.
        """
        self.bottom_toolbar_plugins.append(toolbar)

    def add_sentence_completer(
        self, word_meta_dict: Dict[str, str], ignore_case=True
    ) -> None:
        """Include a custom completer that matches on the entire repl input.

        Args:
            word_meta_dict: Dictionary representing the sentence completions
                and descriptions. Keys are completion text, values are
                descriptions.
        """

        # Don't modify completion if empty.
        if len(word_meta_dict) == 0:
            return

        sentences: List[str] = list(word_meta_dict.keys())
        word_completer = WordCompleter(
            sentences,
            meta_dict=word_meta_dict,
            ignore_case=ignore_case,
            # Whole input field should match
            sentence=True,
        )

        self.extra_completers.append(word_completer)

    def _setup_log_panes(self) -> None:
        """Add loggers to ConsoleApp log pane(s)."""
        if not self.loggers:
            return

        assert isinstance(self.console_app, ConsoleApp)

        if isinstance(self.loggers, list):
            self.console_app.add_log_handler('Logs', self.loggers)

        elif isinstance(self.loggers, dict):
            for window_title, logger_instances in self.loggers.items():
                window_pane = self.console_app.add_log_handler(
                    window_title, logger_instances
                )

                if (
                    window_pane
                    and window_pane.pane_title()
                    in self.hidden_by_default_windows
                ):
                    window_pane.show_pane = False

    def setup_python_logging(
        self,
        last_resort_filename: Optional[str] = None,
        loggers_with_no_propagation: Optional[Iterable[logging.Logger]] = None,
    ) -> None:
        """Setup friendly logging for full-screen prompt_toolkit applications.

        This function sets up Python log handlers to be friendly for full-screen
        prompt_toolkit applications. That is, logging to terminal STDOUT and
        STDERR is disabled so the terminal user interface can be drawn.

        Specifically, all Python STDOUT and STDERR log handlers are
        disabled. It also sets `log propagation to True
        <https://docs.python.org/3/library/logging.html#logging.Logger.propagate>`_.
        to ensure that all log messages are sent to the root logger.

        Args:
            last_resort_filename: If specified use this file as a fallback for
                unhandled Python logging messages. Normally Python will output
                any log messages with no handlers to STDERR as a fallback. If
                None, a temp file will be created instead. See Python
                documentation on `logging.lastResort
                <https://docs.python.org/3/library/logging.html#logging.lastResort>`_
                for more info.
            loggers_with_no_propagation: List of logger instances to skip
               setting ``propagate = True``. This is useful if you would like
               log messages from a particular source to not appear in the root
               logger.
        """
        self.setup_python_logging_called = True
        pw_console_setup_python_logging(
            last_resort_filename, loggers_with_no_propagation
        )

    def hide_windows(self, *window_titles) -> None:
        """Hide window panes specified by title on console startup."""
        for window_title in window_titles:
            self.hidden_by_default_windows.append(window_title)

    def embed(self, override_window_config: Optional[Dict] = None) -> None:
        """Start the console."""

        # Create the ConsoleApp instance.
        self.console_app = ConsoleApp(
            global_vars=self.global_vars,
            local_vars=self.local_vars,
            repl_startup_message=self.repl_startup_message,
            help_text=self.help_text,
            app_title=self.app_title,
            extra_completers=self.extra_completers,
            floating_window_plugins=self.floating_window_plugins,
        )
        PW_CONSOLE_APP_CONTEXTVAR.set(self.console_app)  # type: ignore
        # Setup Python logging and log panes.
        if not self.setup_python_logging_called:
            self.setup_python_logging()
        self._setup_log_panes()

        # Add window pane plugins to the layout.
        for window_pane in self.window_plugins:
            _set_console_app_instance(window_pane, self.console_app)
            # Hide window plugins if the title is hidden by default.
            if window_pane.pane_title() in self.hidden_by_default_windows:
                window_pane.show_pane = False
            self.console_app.window_manager.add_pane(window_pane)

        # Add toolbar plugins to the layout.
        for toolbar in self.top_toolbar_plugins:
            _set_console_app_instance(toolbar, self.console_app)
            self.console_app.window_manager.add_top_toolbar(toolbar)
        for toolbar in self.bottom_toolbar_plugins:
            _set_console_app_instance(toolbar, self.console_app)
            self.console_app.window_manager.add_bottom_toolbar(toolbar)

        # Init floating window plugins.
        for floating_window, _ in self.floating_window_plugins:
            _set_console_app_instance(floating_window, self.console_app)

        # Rebuild prompt_toolkit containers, menu items, and help content with
        # any new plugins added above.
        self.console_app.refresh_layout()

        # Load external config if passed in.
        if self.config_file_path:
            self.console_app.load_clean_config(self.config_file_path)

        if override_window_config:
            self.console_app.prefs.set_windows(override_window_config)
        self.console_app.apply_window_config()

        # Hide the repl pane if it's in the hidden windows list.
        if 'Python Repl' in self.hidden_by_default_windows:
            self.console_app.repl_pane.show_pane = False

        # Start a thread for running user code.
        self.console_app.start_user_code_thread()

        # Startup any background threads and tasks required by plugins.
        for window_pane in self.window_plugins:
            if isinstance(window_pane, PluginMixin):
                window_pane.plugin_start()
        for toolbar in self.bottom_toolbar_plugins:
            if isinstance(toolbar, PluginMixin):
                toolbar.plugin_start()
        for toolbar in self.top_toolbar_plugins:
            if isinstance(toolbar, PluginMixin):
                toolbar.plugin_start()

        # Start the prompt_toolkit UI app.
        asyncio.run(
            self.console_app.run(test_mode=self.test_mode), debug=self.test_mode
        )



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Defines ANSI color codes."""

import ctypes
import os


def _make_color(*codes):
    # Apply all the requested ANSI color codes. Note that this is unbalanced
    # with respect to the reset, which only requires a '0' to erase all codes.
    start = ''.join('\033[{}m'.format(code) for code in codes)
    reset = '\033[0m'

    return staticmethod(lambda msg: u'{}{}{}'.format(start, msg, reset))


class Color:  # pylint: disable=too-few-public-methods
    """Helpers to surround text with ASCII color escapes"""

    bold = _make_color(1)
    red = _make_color(31)
    bold_red = _make_color(31, 1)
    green = _make_color(32)
    magenta = _make_color(35, 1)


def enable_colors():
    if os.name == 'nt':
        kernel32 = ctypes.windll.kernel32
        kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)



#!/usr/bin/env python
# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Generates a BUILD.gn for a Python package.

Pass the script a list of paths to the root directory of a Python package (where
setup.py is).

Don't forget to add the pw_python_package to a top-level group, or it will not
be included in the build.
"""

from datetime import datetime
from pathlib import Path
import sys
from typing import Iterable, List, NamedTuple

from pw_presubmit import git_repo

_HEADER = f"""\
# Copyright {datetime.now().year} The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.

import("//build_overrides/pigweed.gni")

import("$dir_pw_build/python.gni")
"""


class PackageFiles(NamedTuple):
    setup: List[Path]
    sources: List[Path]
    tests: List[Path]
    other: List[Path]


def _find_package_files(root_dir: Path) -> PackageFiles:
    files = git_repo.list_files(
        pathspecs=('*.py', '*.toml', '*.cfg'), repo_path=root_dir
    )

    package_files = PackageFiles([], [], [], [])

    for file in files:
        if file.parent == root_dir:
            if file.name == 'setup.py' or file.suffix != '.py':
                package_files.setup.append(file)
            elif file.stem.startswith('test_') or file.stem.endswith('_test'):
                package_files.tests.append(file)
            else:
                package_files.other.append(file)
        else:
            package_files.sources.append(file)

    return package_files


def _gn_list(name: str, files: Iterable[Path], base: Path) -> Iterable[str]:
    if files:
        yield f'  {name} = ['
        for file in files:
            yield f'    "{file.relative_to(base).as_posix()}",'
        yield '  ]'


def generate_build_gn(root_dir: Path):
    files = _find_package_files(root_dir)

    yield _HEADER

    yield 'pw_python_package("py") {'

    yield from _gn_list('setup', files.setup, root_dir)
    yield from _gn_list('sources', files.sources, root_dir)
    yield from _gn_list('tests', files.tests, root_dir)

    # Don't include the "other" files for now.
    # yield from _gn_list('other', files.other, root_dir)

    yield '}'


def main(paths: Iterable[Path]):
    for path in paths:
        path.joinpath('BUILD.gn').write_text(
            '\n'.join(generate_build_gn(path)) + '\n'
        )


if __name__ == '__main__':
    if len(sys.argv) > 1:
        main(Path(p).resolve() for p in sys.argv[1:])
    else:
        print(__file__, '', __doc__.strip(), sep='\n', file=sys.stderr)
        sys.exit(1)



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Shell related checks."""

import logging
from pw_presubmit import (
    Check,
    PresubmitContext,
    filter_paths,
    tools,
    PresubmitFailure,
)

_LOG = logging.getLogger(__name__)

_SHELL_EXTENSIONS = ('.sh', '.bash')


@filter_paths(endswith=_SHELL_EXTENSIONS)
@Check
def shellcheck(ctx: PresubmitContext) -> None:
    """Run shell script static analiyzer on presubmit."""

    _LOG.warning(
        "The Pigweed project discourages use of shellscripts. "
        "https://pigweed.dev/docs/faq.html"
    )

    result = tools.log_run(['shellcheck', *ctx.paths])
    if result.returncode != 0:
        raise PresubmitFailure('Shellcheck identifed issues.')



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: pw_unit_test_proto/unit_test.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='pw_unit_test_proto/unit_test.proto',
  package='pw.unit_test',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\"pw_unit_test_proto/unit_test.proto\x12\x0cpw.unit_test\"N\n\x12TestCaseDescriptor\x12\x12\n\nsuite_name\x18\x01 \x01(\t\x12\x11\n\ttest_name\x18\x02 \x01(\t\x12\x11\n\tfile_name\x18\x03 \x01(\t\"m\n\x13TestCaseExpectation\x12\x12\n\nexpression\x18\x01 \x01(\t\x12\x1c\n\x14\x65valuated_expression\x18\x02 \x01(\t\x12\x13\n\x0bline_number\x18\x03 \x01(\r\x12\x0f\n\x07success\x18\x04 \x01(\x08\"\x0e\n\x0cTestRunStart\"O\n\nTestRunEnd\x12\x0e\n\x06passed\x18\x01 \x01(\r\x12\x0e\n\x06\x66\x61iled\x18\x02 \x01(\r\x12\x0f\n\x07skipped\x18\x03 \x01(\r\x12\x10\n\x08\x64isabled\x18\x04 \x01(\r\"\xef\x02\n\x05\x45vent\x12\x34\n\x0etest_run_start\x18\x01 \x01(\x0b\x32\x1a.pw.unit_test.TestRunStartH\x00\x12\x30\n\x0ctest_run_end\x18\x02 \x01(\x0b\x32\x18.pw.unit_test.TestRunEndH\x00\x12;\n\x0ftest_case_start\x18\x03 \x01(\x0b\x32 .pw.unit_test.TestCaseDescriptorH\x00\x12\x35\n\rtest_case_end\x18\x04 \x01(\x0e\x32\x1c.pw.unit_test.TestCaseResultH\x00\x12>\n\x12test_case_disabled\x18\x05 \x01(\x0b\x32 .pw.unit_test.TestCaseDescriptorH\x00\x12\x42\n\x15test_case_expectation\x18\x06 \x01(\x0b\x32!.pw.unit_test.TestCaseExpectationH\x00\x42\x06\n\x04type\"H\n\x0eTestRunRequest\x12\"\n\x1areport_passed_expectations\x18\x01 \x01(\x08\x12\x12\n\ntest_suite\x18\x02 \x03(\t*7\n\x0eTestCaseResult\x12\x0b\n\x07SUCCESS\x10\x00\x12\x0b\n\x07\x46\x41ILURE\x10\x01\x12\x0b\n\x07SKIPPED\x10\x02\x32H\n\x08UnitTest\x12<\n\x03Run\x12\x1c.pw.unit_test.TestRunRequest\x1a\x13.pw.unit_test.Event\"\x00\x30\x01\x62\x06proto3'
)

_TESTCASERESULT = _descriptor.EnumDescriptor(
  name='TestCaseResult',
  full_name='pw.unit_test.TestCaseResult',
  filename=None,
  file=DESCRIPTOR,
  create_key=_descriptor._internal_create_key,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SUCCESS', index=0, number=0,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='FAILURE', index=1, number=1,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
    _descriptor.EnumValueDescriptor(
      name='SKIPPED', index=2, number=2,
      serialized_options=None,
      type=None,
      create_key=_descriptor._internal_create_key),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=784,
  serialized_end=839,
)
_sym_db.RegisterEnumDescriptor(_TESTCASERESULT)

TestCaseResult = enum_type_wrapper.EnumTypeWrapper(_TESTCASERESULT)
SUCCESS = 0
FAILURE = 1
SKIPPED = 2



_TESTCASEDESCRIPTOR = _descriptor.Descriptor(
  name='TestCaseDescriptor',
  full_name='pw.unit_test.TestCaseDescriptor',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='suite_name', full_name='pw.unit_test.TestCaseDescriptor.suite_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='test_name', full_name='pw.unit_test.TestCaseDescriptor.test_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='file_name', full_name='pw.unit_test.TestCaseDescriptor.file_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=52,
  serialized_end=130,
)


_TESTCASEEXPECTATION = _descriptor.Descriptor(
  name='TestCaseExpectation',
  full_name='pw.unit_test.TestCaseExpectation',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='expression', full_name='pw.unit_test.TestCaseExpectation.expression', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='evaluated_expression', full_name='pw.unit_test.TestCaseExpectation.evaluated_expression', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=b"".decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='line_number', full_name='pw.unit_test.TestCaseExpectation.line_number', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='success', full_name='pw.unit_test.TestCaseExpectation.success', index=3,
      number=4, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=132,
  serialized_end=241,
)


_TESTRUNSTART = _descriptor.Descriptor(
  name='TestRunStart',
  full_name='pw.unit_test.TestRunStart',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=243,
  serialized_end=257,
)


_TESTRUNEND = _descriptor.Descriptor(
  name='TestRunEnd',
  full_name='pw.unit_test.TestRunEnd',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='passed', full_name='pw.unit_test.TestRunEnd.passed', index=0,
      number=1, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='failed', full_name='pw.unit_test.TestRunEnd.failed', index=1,
      number=2, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='skipped', full_name='pw.unit_test.TestRunEnd.skipped', index=2,
      number=3, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='disabled', full_name='pw.unit_test.TestRunEnd.disabled', index=3,
      number=4, type=13, cpp_type=3, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=259,
  serialized_end=338,
)


_EVENT = _descriptor.Descriptor(
  name='Event',
  full_name='pw.unit_test.Event',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='test_run_start', full_name='pw.unit_test.Event.test_run_start', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='test_run_end', full_name='pw.unit_test.Event.test_run_end', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='test_case_start', full_name='pw.unit_test.Event.test_case_start', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='test_case_end', full_name='pw.unit_test.Event.test_case_end', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='test_case_disabled', full_name='pw.unit_test.Event.test_case_disabled', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='test_case_expectation', full_name='pw.unit_test.Event.test_case_expectation', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='type', full_name='pw.unit_test.Event.type',
      index=0, containing_type=None,
      create_key=_descriptor._internal_create_key,
    fields=[]),
  ],
  serialized_start=341,
  serialized_end=708,
)


_TESTRUNREQUEST = _descriptor.Descriptor(
  name='TestRunRequest',
  full_name='pw.unit_test.TestRunRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='report_passed_expectations', full_name='pw.unit_test.TestRunRequest.report_passed_expectations', index=0,
      number=1, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='test_suite', full_name='pw.unit_test.TestRunRequest.test_suite', index=1,
      number=2, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=710,
  serialized_end=782,
)

_EVENT.fields_by_name['test_run_start'].message_type = _TESTRUNSTART
_EVENT.fields_by_name['test_run_end'].message_type = _TESTRUNEND
_EVENT.fields_by_name['test_case_start'].message_type = _TESTCASEDESCRIPTOR
_EVENT.fields_by_name['test_case_end'].enum_type = _TESTCASERESULT
_EVENT.fields_by_name['test_case_disabled'].message_type = _TESTCASEDESCRIPTOR
_EVENT.fields_by_name['test_case_expectation'].message_type = _TESTCASEEXPECTATION
_EVENT.oneofs_by_name['type'].fields.append(
  _EVENT.fields_by_name['test_run_start'])
_EVENT.fields_by_name['test_run_start'].containing_oneof = _EVENT.oneofs_by_name['type']
_EVENT.oneofs_by_name['type'].fields.append(
  _EVENT.fields_by_name['test_run_end'])
_EVENT.fields_by_name['test_run_end'].containing_oneof = _EVENT.oneofs_by_name['type']
_EVENT.oneofs_by_name['type'].fields.append(
  _EVENT.fields_by_name['test_case_start'])
_EVENT.fields_by_name['test_case_start'].containing_oneof = _EVENT.oneofs_by_name['type']
_EVENT.oneofs_by_name['type'].fields.append(
  _EVENT.fields_by_name['test_case_end'])
_EVENT.fields_by_name['test_case_end'].containing_oneof = _EVENT.oneofs_by_name['type']
_EVENT.oneofs_by_name['type'].fields.append(
  _EVENT.fields_by_name['test_case_disabled'])
_EVENT.fields_by_name['test_case_disabled'].containing_oneof = _EVENT.oneofs_by_name['type']
_EVENT.oneofs_by_name['type'].fields.append(
  _EVENT.fields_by_name['test_case_expectation'])
_EVENT.fields_by_name['test_case_expectation'].containing_oneof = _EVENT.oneofs_by_name['type']
DESCRIPTOR.message_types_by_name['TestCaseDescriptor'] = _TESTCASEDESCRIPTOR
DESCRIPTOR.message_types_by_name['TestCaseExpectation'] = _TESTCASEEXPECTATION
DESCRIPTOR.message_types_by_name['TestRunStart'] = _TESTRUNSTART
DESCRIPTOR.message_types_by_name['TestRunEnd'] = _TESTRUNEND
DESCRIPTOR.message_types_by_name['Event'] = _EVENT
DESCRIPTOR.message_types_by_name['TestRunRequest'] = _TESTRUNREQUEST
DESCRIPTOR.enum_types_by_name['TestCaseResult'] = _TESTCASERESULT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

TestCaseDescriptor = _reflection.GeneratedProtocolMessageType('TestCaseDescriptor', (_message.Message,), {
  'DESCRIPTOR' : _TESTCASEDESCRIPTOR,
  '__module__' : 'pw_unit_test_proto.unit_test_pb2'
  # @@protoc_insertion_point(class_scope:pw.unit_test.TestCaseDescriptor)
  })
_sym_db.RegisterMessage(TestCaseDescriptor)

TestCaseExpectation = _reflection.GeneratedProtocolMessageType('TestCaseExpectation', (_message.Message,), {
  'DESCRIPTOR' : _TESTCASEEXPECTATION,
  '__module__' : 'pw_unit_test_proto.unit_test_pb2'
  # @@protoc_insertion_point(class_scope:pw.unit_test.TestCaseExpectation)
  })
_sym_db.RegisterMessage(TestCaseExpectation)

TestRunStart = _reflection.GeneratedProtocolMessageType('TestRunStart', (_message.Message,), {
  'DESCRIPTOR' : _TESTRUNSTART,
  '__module__' : 'pw_unit_test_proto.unit_test_pb2'
  # @@protoc_insertion_point(class_scope:pw.unit_test.TestRunStart)
  })
_sym_db.RegisterMessage(TestRunStart)

TestRunEnd = _reflection.GeneratedProtocolMessageType('TestRunEnd', (_message.Message,), {
  'DESCRIPTOR' : _TESTRUNEND,
  '__module__' : 'pw_unit_test_proto.unit_test_pb2'
  # @@protoc_insertion_point(class_scope:pw.unit_test.TestRunEnd)
  })
_sym_db.RegisterMessage(TestRunEnd)

Event = _reflection.GeneratedProtocolMessageType('Event', (_message.Message,), {
  'DESCRIPTOR' : _EVENT,
  '__module__' : 'pw_unit_test_proto.unit_test_pb2'
  # @@protoc_insertion_point(class_scope:pw.unit_test.Event)
  })
_sym_db.RegisterMessage(Event)

TestRunRequest = _reflection.GeneratedProtocolMessageType('TestRunRequest', (_message.Message,), {
  'DESCRIPTOR' : _TESTRUNREQUEST,
  '__module__' : 'pw_unit_test_proto.unit_test_pb2'
  # @@protoc_insertion_point(class_scope:pw.unit_test.TestRunRequest)
  })
_sym_db.RegisterMessage(TestRunRequest)



_UNITTEST = _descriptor.ServiceDescriptor(
  name='UnitTest',
  full_name='pw.unit_test.UnitTest',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=841,
  serialized_end=913,
  methods=[
  _descriptor.MethodDescriptor(
    name='Run',
    full_name='pw.unit_test.UnitTest.Run',
    index=0,
    containing_service=None,
    input_type=_TESTRUNREQUEST,
    output_type=_EVENT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_UNITTEST)

DESCRIPTOR.services_by_name['UnitTest'] = _UNITTEST

# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-

# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Prints the env_setup banner for cmd.exe.

This is done from Python as activating colors and printing ASCII art are not
easy to do in cmd.exe. Activated colors also don't persist in the parent
process.
"""

from __future__ import print_function

import argparse
import os
import sys

from .colors import Color, enable_colors  # type: ignore

_PIGWEED_BANNER = u'''
 ▒█████▄   █▓  ▄███▒  ▒█    ▒█ ░▓████▒ ░▓████▒ ▒▓████▄
  ▒█░  █░ ░█▒ ██▒ ▀█▒ ▒█░ █ ▒█  ▒█   ▀  ▒█   ▀  ▒█  ▀█▌
  ▒█▄▄▄█░ ░█▒ █▓░ ▄▄░ ▒█░ █ ▒█  ▒███    ▒███    ░█   █▌
  ▒█▀     ░█░ ▓█   █▓ ░█░ █ ▒█  ▒█   ▄  ▒█   ▄  ░█  ▄█▌
  ▒█      ░█░ ░▓███▀   ▒█▓▀▓█░ ░▓████▒ ░▓████▒ ▒▓████▀
'''


def print_banner(bootstrap, no_shell_file):
    """Print the Pigweed or project-specific banner"""
    enable_colors()

    print(Color.green('\n  WELCOME TO...'))
    print(Color.magenta(_PIGWEED_BANNER))

    if bootstrap:
        print(
            Color.green(
                '\n  BOOTSTRAP! Bootstrap may take a few minutes; '
                'please be patient'
            )
        )
        print(
            Color.green(
                '  On Windows, this stage is extremely slow (~10 minutes).\n'
            )
        )
    else:
        print(
            Color.green(
                '\n  ACTIVATOR! This sets your console environment variables.\n'
            )
        )

        if no_shell_file:
            print(Color.bold_red('Error!\n'))
            print(
                Color.red(
                    '  Your Pigweed environment does not seem to be'
                    ' configured.'
                )
            )
            print(Color.red('  Run bootstrap.bat to perform initial setup.'))

    return 0


def parse():
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        prog="python -m pw_env_setup.windows_env_start"
    )
    parser.add_argument('--bootstrap', action='store_true')
    parser.add_argument('--no-shell-file', action='store_true')
    return parser.parse_args()


def main():
    """Script entry point."""
    if os.name != 'nt':
        return 1
    return print_banner(**vars(parse()))


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3
# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Launch a pw_target_runner server to use for multi-device testing."""

import argparse
import logging
import sys
import tempfile
from typing import IO, List, Optional

import pw_cli.process

import pw_arduino_build.log
from pw_arduino_build import teensy_detector
from pw_arduino_build.file_operations import decode_file_json
from pw_arduino_build.unit_test_runner import ArduinoCoreNotSupported

_LOG = logging.getLogger('unit_test_server')

_TEST_RUNNER_COMMAND = 'arduino_unit_test_runner'

_TEST_SERVER_COMMAND = 'pw_target_runner_server'


class UnknownArduinoCore(Exception):
    """Exception raised when no Arduino core can be found."""


def parse_args():
    """Parses command-line arguments."""

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        '--server-port',
        type=int,
        default=8081,
        help='Port to launch the pw_target_runner_server on',
    )
    parser.add_argument(
        '--server-config',
        type=argparse.FileType('r'),
        help='Path to server config file',
    )
    parser.add_argument(
        '--verbose',
        '-v',
        dest='verbose',
        action="store_true",
        help='Output additional logs as the script runs',
    )
    parser.add_argument(
        "-c",
        "--config-file",
        required=True,
        help="Path to an arduino_builder config file.",
    )
    # TODO(tonymd): Explicitly split args using "--". See example in:
    # //pw_unit_test/py/pw_unit_test/test_runner.py:326
    parser.add_argument(
        'runner_args',
        nargs=argparse.REMAINDER,
        help='Arguments to forward to the test runner',
    )

    return parser.parse_args()


def generate_runner(command: str, arguments: List[str]) -> str:
    """Generates a text-proto style pw_target_runner_server configuration."""
    # TODO(amontanez): Use a real proto library to generate this when we have
    # one set up.
    for i, arg in enumerate(arguments):
        arguments[i] = f'  args: "{arg}"'
    runner = ['runner {', f'  command:"{command}"']
    runner.extend(arguments)
    runner.append('}\n')
    return '\n'.join(runner)


def generate_server_config(
    runner_args: Optional[List[str]], arduino_package_path: str
) -> IO[bytes]:
    """Returns a temporary generated file for use as the server config."""

    if "teensy" not in arduino_package_path:
        raise ArduinoCoreNotSupported(arduino_package_path)

    boards = teensy_detector.detect_boards(arduino_package_path)
    if not boards:
        _LOG.critical('No attached boards detected')
        sys.exit(1)
    config_file = tempfile.NamedTemporaryFile()
    _LOG.debug('Generating test server config at %s', config_file.name)
    _LOG.debug('Found %d attached devices', len(boards))
    for board in boards:
        test_runner_args = []
        if runner_args:
            test_runner_args += runner_args
        test_runner_args += ["-v"] + board.test_runner_args()
        test_runner_args += ["--port", board.dev_name]
        test_runner_args += ["--upload-tool", board.arduino_upload_tool_name]
        config_file.write(
            generate_runner(_TEST_RUNNER_COMMAND, test_runner_args).encode(
                'utf-8'
            )
        )
    config_file.flush()
    return config_file


def launch_server(
    server_config: Optional[IO[bytes]],
    server_port: Optional[int],
    runner_args: Optional[List[str]],
    arduino_package_path: str,
) -> int:
    """Launch a device test server with the provided arguments."""
    if server_config is None:
        # Auto-detect attached boards if no config is provided.
        server_config = generate_server_config(
            runner_args, arduino_package_path
        )

    cmd = [_TEST_SERVER_COMMAND, '-config', server_config.name]

    if server_port is not None:
        cmd.extend(['-port', str(server_port)])

    return pw_cli.process.run(*cmd, log_output=True).returncode


def main():
    """Launch a device test server with the provided arguments."""
    args = parse_args()

    if "--" in args.runner_args:
        args.runner_args.remove("--")

    log_level = logging.DEBUG if args.verbose else logging.INFO
    pw_arduino_build.log.install(log_level)

    # Get arduino_package_path from either the config file or command line args.
    arduino_package_path = None
    if args.config_file:
        json_file_options, unused_config_path = decode_file_json(
            args.config_file
        )
        arduino_package_path = json_file_options.get(
            "arduino_package_path", None
        )
        # Must pass --config-file option in the runner_args.
        if "--config-file" not in args.runner_args:
            args.runner_args.append("--config-file")
            args.runner_args.append(args.config_file)

    # Check for arduino_package_path in the runner_args
    try:
        arduino_package_path = args.runner_args[
            args.runner_args.index("--arduino-package-path") + 1
        ]
    except (ValueError, IndexError):
        # Only raise an error if arduino_package_path not set from the json.
        if arduino_package_path is None:
            raise UnknownArduinoCore(
                "Test runner arguments: '{}'".format(" ".join(args.runner_args))
            )

    exit_code = launch_server(
        args.server_config,
        args.server_port,
        args.runner_args,
        arduino_package_path,
    )
    sys.exit(exit_code)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Installs or updates prebuilt tools.

Must be tested with Python 2 and Python 3.

The stdout of this script is meant to be executed by the invoking shell.
"""

from __future__ import print_function

import collections
import hashlib
import json
import os
import platform as platform_module
import re
import subprocess
import sys


def check_auth(cipd, package_files, cipd_service_account, spin):
    """Check have access to CIPD pigweed directory."""
    cmd = [cipd]
    extra_args = []
    if cipd_service_account:
        extra_args.extend(['-service-account-json', cipd_service_account])

    paths = []
    for package_file in package_files:
        with open(package_file, 'r') as ins:
            # This is an expensive RPC, so only check the first few entries
            # in each file.
            for i, entry in enumerate(json.load(ins).get('packages', ())):
                if i >= 3:
                    break
                parts = entry['path'].split('/')
                while '${' in parts[-1]:
                    parts.pop(-1)
                paths.append('/'.join(parts))

    username = None
    try:
        output = subprocess.check_output(
            cmd + ['auth-info'] + extra_args, stderr=subprocess.STDOUT
        ).decode()
        logged_in = True

        match = re.search(r'Logged in as (\S*)\.', output)
        if match:
            username = match.group(1)

    except subprocess.CalledProcessError:
        logged_in = False

    def _check_all_paths():
        inaccessible_paths = []

        for path in paths:
            # Not catching CalledProcessError because 'cipd ls' seems to never
            # return an error code unless it can't reach the CIPD server.
            output = subprocess.check_output(
                cmd + ['ls', path] + extra_args, stderr=subprocess.STDOUT
            ).decode()
            if 'No matching packages' not in output:
                continue

            # 'cipd ls' only lists sub-packages but ignores any packages at the
            # given path. 'cipd instances' will give versions of that package.
            # 'cipd instances' does use an error code if there's no such package
            # or that package is inaccessible.
            try:
                subprocess.check_output(
                    cmd + ['instances', path] + extra_args,
                    stderr=subprocess.STDOUT,
                )
            except subprocess.CalledProcessError:
                inaccessible_paths.append(path)

        return inaccessible_paths

    inaccessible_paths = _check_all_paths()

    if inaccessible_paths and not logged_in:
        with spin.pause():
            stderr = lambda *args: print(*args, file=sys.stderr)
            stderr()
            stderr(
                'Not logged in to CIPD and no anonymous access to the '
                'following CIPD paths:'
            )
            for path in inaccessible_paths:
                stderr('  {}'.format(path))
            stderr()
            stderr('Attempting CIPD login')
            try:
                # Note that with -service-account-json, auth-login is a no-op.
                subprocess.check_call(cmd + ['auth-login'] + extra_args)
            except subprocess.CalledProcessError:
                stderr('CIPD login failed')
                return False

        inaccessible_paths = _check_all_paths()

    if inaccessible_paths:
        stderr = lambda *args: print(*args, file=sys.stderr)
        stderr('=' * 60)
        username_part = ''
        if username:
            username_part = '({}) '.format(username)
        stderr(
            'Your account {}does not have access to the following '
            'paths'.format(username_part)
        )
        stderr('(or they do not exist)')
        for path in inaccessible_paths:
            stderr('  {}'.format(path))
        stderr('=' * 60)
        return False

    return True


def platform(rosetta=False):
    """Return the CIPD platform string of the current system."""
    osname = {
        'darwin': 'mac',
        'linux': 'linux',
        'windows': 'windows',
    }[platform_module.system().lower()]

    if platform_module.machine().startswith(('aarch64', 'armv8')):
        arch = 'arm64'
    elif platform_module.machine() == 'x86_64':
        arch = 'amd64'
    elif platform_module.machine() == 'i686':
        arch = 'i386'
    else:
        arch = platform_module.machine()

    platform_arch = '{}-{}'.format(osname, arch).lower()

    # Support `mac-arm64` through Rosetta until `mac-arm64` binaries are ready
    if platform_arch == 'mac-arm64' and rosetta:
        return 'mac-amd64'

    return platform_arch


def all_package_files(env_vars, package_files):
    """Recursively retrieve all package files."""

    result = []
    to_process = []
    for pkg_file in package_files:
        args = []
        if env_vars:
            args.append(env_vars.get('PW_PROJECT_ROOT'))
        args.append(pkg_file)

        # The signature here is os.path.join(a, *p). Pylint doesn't like when
        # we call os.path.join(*args), but is happy if we instead call
        # os.path.join(args[0], *args[1:]). Disabling the option on this line
        # seems to be a less confusing choice.
        path = os.path.join(*args)  # pylint: disable=no-value-for-parameter

        to_process.append(path)

    while to_process:
        package_file = to_process.pop(0)
        result.append(package_file)

        with open(package_file, 'r') as ins:
            entries = json.load(ins).get('included_files', ())

        for entry in entries:
            entry = os.path.join(os.path.dirname(package_file), entry)

            if entry not in result and entry not in to_process:
                to_process.append(entry)

    return result


def all_packages(package_files):
    packages = []
    for package_file in package_files:
        name = package_file_name(package_file)
        with open(package_file, 'r') as ins:
            file_packages = json.load(ins).get('packages', ())
            for package in file_packages:
                if 'subdir' in package:
                    package['subdir'] = os.path.join(name, package['subdir'])
                else:
                    package['subdir'] = name
            packages.extend(file_packages)
    return packages


def deduplicate_packages(packages):
    deduped = collections.OrderedDict()
    for package in reversed(packages):
        if package['path'] in deduped:
            del deduped[package['path']]
        deduped[package['path']] = package
    return reversed(list(deduped.values()))


def write_ensure_file(
    package_files, ensure_file, platform
):  # pylint: disable=redefined-outer-name
    packages = all_packages(package_files)
    deduped_packages = deduplicate_packages(packages)

    with open(ensure_file, 'w') as outs:
        outs.write(
            '$VerifiedPlatform linux-amd64\n'
            '$VerifiedPlatform mac-amd64\n'
            '$ParanoidMode CheckPresence\n'
        )

        for pkg in deduped_packages:
            # If this is a new-style package manifest platform handling must
            # be done here instead of by the cipd executable.
            if 'platforms' in pkg and platform not in pkg['platforms']:
                continue

            outs.write('@Subdir {}\n'.format(pkg.get('subdir', '')))
            outs.write('{} {}\n'.format(pkg['path'], ' '.join(pkg['tags'])))


def package_file_name(package_file):
    return os.path.basename(os.path.splitext(package_file)[0])


def package_installation_path(root_install_dir, package_file):
    """Returns the package installation path.

    Args:
      root_install_dir: The CIPD installation directory.
      package_file: The path to the .json package definition file.
    """
    return os.path.join(
        root_install_dir, 'packages', package_file_name(package_file)
    )


def update(  # pylint: disable=too-many-locals
    cipd,
    package_files,
    root_install_dir,
    cache_dir,
    rosetta=False,
    env_vars=None,
    spin=None,
    trust_hash=False,
):
    """Grab the tools listed in ensure_files."""

    package_files = all_package_files(env_vars, package_files)

    # TODO(mohrr) use os.makedirs(..., exist_ok=True).
    if not os.path.isdir(root_install_dir):
        os.makedirs(root_install_dir)

    # This file is read by 'pw doctor' which needs to know which package files
    # were used in the environment.
    package_files_file = os.path.join(
        root_install_dir, '_all_package_files.json'
    )
    with open(package_files_file, 'w') as outs:
        json.dump(package_files, outs, indent=2)

    if env_vars:
        env_vars.prepend('PATH', root_install_dir)
        env_vars.set('PW_CIPD_INSTALL_DIR', root_install_dir)
        env_vars.set('CIPD_CACHE_DIR', cache_dir)

    pw_root = None

    if env_vars:
        pw_root = env_vars.get('PW_ROOT', None)
    if not pw_root:
        pw_root = os.environ['PW_ROOT']

    plat = platform(rosetta)

    ensure_file = os.path.join(root_install_dir, 'packages.ensure')
    write_ensure_file(package_files, ensure_file, plat)

    install_dir = os.path.join(root_install_dir, 'packages')

    cmd = [
        cipd,
        'ensure',
        '-ensure-file',
        ensure_file,
        '-root',
        install_dir,
        '-log-level',
        'debug',
        '-json-output',
        os.path.join(root_install_dir, 'packages.json'),
        '-cache-dir',
        cache_dir,
        '-max-threads',
        '0',  # 0 means use CPU count.
    ]

    cipd_service_account = None
    if env_vars:
        cipd_service_account = env_vars.get('PW_CIPD_SERVICE_ACCOUNT_JSON')
    if not cipd_service_account:
        cipd_service_account = os.environ.get('PW_CIPD_SERVICE_ACCOUNT_JSON')
    if cipd_service_account:
        cmd.extend(['-service-account-json', cipd_service_account])

    hasher = hashlib.sha256()
    encoded = '\0'.join(cmd)
    if hasattr(encoded, 'encode'):
        encoded = encoded.encode()
    hasher.update(encoded)
    with open(ensure_file, 'rb') as ins:
        hasher.update(ins.read())
    digest = hasher.hexdigest()

    with open(os.path.join(root_install_dir, 'hash.log'), 'w') as hashlog:
        print('calculated digest:', digest, file=hashlog)

        hash_file = os.path.join(root_install_dir, 'packages.sha256')
        print('hash file path:', hash_file, file=hashlog)
        print('exists:', os.path.isfile(hash_file), file=hashlog)
        print('trust_hash:', trust_hash, file=hashlog)
        if trust_hash and os.path.isfile(hash_file):
            with open(hash_file, 'r') as ins:
                digest_file = ins.read().strip()
                print('contents:', digest_file, file=hashlog)
                print('equal:', digest == digest_file, file=hashlog)
                if digest == digest_file:
                    return True

    if not check_auth(cipd, package_files, cipd_service_account, spin):
        return False

    log = os.path.join(root_install_dir, 'packages.log')
    try:
        with open(log, 'w') as outs:
            print(*cmd, file=outs)
            subprocess.check_call(cmd, stdout=outs, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError:
        with open(log, 'r') as ins:
            sys.stderr.write(ins.read())
            raise

    with open(hash_file, 'w') as outs:
        print(digest, file=outs)

    # Set environment variables so tools can later find things under, for
    # example, 'share'.
    if env_vars:
        for package_file in reversed(package_files):
            name = package_file_name(package_file)
            file_install_dir = os.path.join(install_dir, name)
            # Some executables get installed at top-level and some get
            # installed under 'bin'. A small number of old packages prefix the
            # entire tree with the platform (e.g., chromium/third_party/tcl).
            for bin_dir in (
                file_install_dir,
                os.path.join(file_install_dir, 'bin'),
                os.path.join(file_install_dir, plat, 'bin'),
            ):
                if os.path.isdir(bin_dir):
                    env_vars.prepend('PATH', bin_dir)
            env_vars.set(
                'PW_{}_CIPD_INSTALL_DIR'.format(name.upper()), file_install_dir
            )

            # Windows has its own special toolchain.
            if os.name == 'nt':
                env_vars.prepend(
                    'PATH', os.path.join(file_install_dir, 'mingw64', 'bin')
                )

    return True
